/*--------------------------------------------------------------*/
/* helpme.c:							*/
/*	Primary routines for the "helpme" game			*/
/* Copyright (c) 2009  Tim Edwards, Open Circuit Design		*/
/*--------------------------------------------------------------*/

#include <stdio.h>
#include <stdarg.h>	/* for va_copy() */
#include <stdlib.h>	/* for atoi() and others */
#include <ctype.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <errno.h>
#include <math.h>

#include "helpme.h"

/*--------------------------------------------------------------*/
/* Compare a (filled) timeval struct against the one used by	*/
/* "now" to track current (real) time.  Return 1 if "check" is	*/
/* earlier than "now", 0 if later.				*/
/*--------------------------------------------------------------*/

int compare_time(struct timeval *timeout)
{
   if (timeout->tv_sec < hms.now.tv_sec) return 1;
   else if (timeout->tv_sec > hms.now.tv_sec) return 0;
   else if (timeout->tv_usec < hms.now.tv_usec) return 1;
   else return 0;
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void set_future_time(struct timeval *timeout, float seconds)
{
   int isec = (int)seconds;
   int usec = (int)((seconds - (float)isec) * 1000000);

   timeout->tv_sec = hms.now.tv_sec + isec;
   timeout->tv_usec = hms.now.tv_usec + usec;
   if (timeout->tv_usec >= 1000000) {
      timeout->tv_sec++;
      timeout->tv_usec -= 1000000;
   }
}

/*--------------------------------------------------------------*/
/* Force an angle value in degrees to be between (-180, 180)	*/
/*--------------------------------------------------------------*/

void normalize_angle(int *angle)
{
   if (*angle > 1800) *angle -= 3600;
   else if (*angle < -1800) *angle += 3600;
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void get_field_position(int x, int y, fpoint *fpout)
{
   fpoint fpin;
   Matrix *mat;

   fpin.x = (float)x;
   fpin.y = (float)y;

   mat = &hms.screen2field;
   fpin.z = (-fpin.x * mat->c - fpin.y * mat->f - mat->l) / mat->i;

   MultiplyByMatrix(mat, &fpin, fpout);
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void get_screen_position(fpoint *fpin, int *xout, int *yout)
{
   fpoint fpout;

   MultiplyByMatrix(&hms.field2screen, fpin, &fpout);

   *xout = (int)fpout.x;
   *yout = (int)fpout.y;
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

#define NORTH   0x01
#define SOUTH   0x02
#define EAST    0x04
#define WEST    0x08
#define VISITED 0x10

u_char expand_bounds(int x, int y, u_char *list, int maxx, int maxy)
{
   int entry;

   /* Check out-of-bounds */
   if ((x < 0) || (x > maxx) || (y < 0) || (y > maxy)) return 0;

   /* Check for area already visited */
   entry = y * (maxx + 1) + x;
   if (list[entry] & VISITED) return 1;

   /* Mark area as visited */
   list[entry] |= VISITED;

   if ((list[entry] & NORTH) == 0) {
      if (expand_bounds(x, y + 1, list, maxx, maxy) == 0) return 0;
   }
   if ((list[entry] & EAST) == 0) {
      if (expand_bounds(x + 1, y, list, maxx, maxy) == 0) return 0;
   }
   if ((list[entry] & SOUTH) == 0) {
      if (expand_bounds(x, y - 1, list, maxx, maxy) == 0) return 0;
   }
   if ((list[entry] & WEST) == 0) {
      if (expand_bounds(x - 1, y, list, maxx, maxy) == 0) return 0;
   }
   return 1;	/* Continue searching */
}

/*--------------------------------------------------------------*/
/* Check conditions for victory:				*/
/* (1) Castle is complete---forms an enclosed structure		*/
/* (2) Catapult is inside an enclosed structure			*/
/* (3) There are no bomb targets				*/
/* (4) No enemy soldiers have bombs				*/
/*--------------------------------------------------------------*/

u_char check_victory()
{
    u_char rval, *list;
    int i, minx, maxx, miny, maxy, x, y, cx, cy, entry;
    int tarea, trows, tcols;
    fpoint fp;
    Castle *wall;
    Target *target;
    Enemy *enemy;

    for (i = 0; i < hms.num_targets; i++) {
       target = hms.targets + i;
       if (target->state != GONE && target->type == BOMB)
	  return 0;
    }

    for (i = 0; i < hms.num_enemies; i++) {
       enemy = hms.enemies + i;
       if (enemy->flags == BOMB)
	  return 0;
    }

    /* Castle boundary scan */
    if (hms.num_pieces == 0) return 0;

    /* Find wall minimum and maximum values */
    minx = miny = 1000;
    maxx = maxy = -1000;
    for (i = 0; i < hms.num_pieces; i++) {
        wall = hms.pieces + i;

        x = ((int)(wall->pos[1].x + 10025.0) / 50) - 200;
        y = ((int)(wall->pos[1].y + 10025.0) / 50) - 200;

	if (x < minx) minx = x;
	if ((wall->align == XALIGNED) && (x > maxx)) maxx = x;
	if (y < miny) miny = y;
	if ((wall->align == YALIGNED) && (y > maxy)) maxy = y;
    }

    cx = ((int)(hms.catapult->pos.x + 10025.0) / 50) - 200;
    cy = ((int)(hms.catapult->pos.y + 10025.0) / 50) - 200;

    /* Check out-of-bounds */
    if ((cx < minx) || (cx > (maxx + 1)) || (cy < miny) || (cy > (maxy + 1)))
       return 0;

    /* Generate matrix of the area inside the walls */

    tcols = (maxx - minx + 1);
    trows = (maxy - miny + 1);
    tarea = tcols * trows;
    list = (u_char *)Tcl_Alloc(tarea * sizeof(u_char));
    memset((void *)list, 0, tarea);

    cx -= minx;
    cy -= miny;
    maxx -= minx;
    maxy -= miny;

    /* Fill the matrix with blocking information */

    for (i = 0; i < hms.num_pieces; i++) {
       wall = hms.pieces + i;
       x = ((int)(wall->pos[1].x + 10025.0) / 50) - 200;
       y = ((int)(wall->pos[1].y + 10025.0) / 50) - 200;
       x -= minx;
       y -= miny;
       entry = x + tcols * y;

       if (wall->align == XALIGNED) {
	  if (y > 0) list[entry - tcols] |= NORTH;
	  if (y <= maxy) list[entry] |= SOUTH;
       }
       else {
	  if (x > 0) list[entry - 1] |= EAST;
	  if (x <= maxx) list[entry] |= WEST;
       }
    }

    /* Recursively mark the matrix until we are blocked on all	*/
    /* sides (victory) or reach one of the bounds (failure)	*/

    rval = expand_bounds(cx, cy, list, maxx, maxy);
    free(list);
    return rval;
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void calc_arm_pos(fpoint *new_pos)
{
    float arm_angle, tmpx;

    // Compute catapult arm position;  set boulder position at
    // the end of the arm.  Angle of the arm is 50 degrees at
    // pull = 0 to 0 degrees at pull = 500.

    arm_angle = (400 + hms.catapult->pull) * DEG_TO_RAD;
    tmpx =  -30 * sin(arm_angle);
    new_pos->z =  30 * cos(arm_angle);

    new_pos->x = tmpx * cos(hms.catapult->angle * DEG_TO_RAD); 
    new_pos->y = tmpx * sin(hms.catapult->angle * DEG_TO_RAD); 

    new_pos->x += hms.catapult->pos.x;
    new_pos->y += hms.catapult->pos.y;
    new_pos->z += hms.catapult->pos.z;
}

/*--------------------------------------------------------------*/
/* Compute the catapult corner positions			*/
/* Return 1 if position collides with a wall, 0 if not.		*/			
/*--------------------------------------------------------------*/

u_char catapult_corners()
{
   fpoint catapult_vec[4] = {
	{ 30.0,  20.0,  0.0},
	{ 30.0, -20.0,  0.0},
	{-30.0, -20.0,  0.0},
	{-30.0,  20.0,  0.0}
   };

   int i, j;
   fpoint *fptptr;
   float csin, ccos, d;
   Catapult *catapult = hms.catapult;
   Castle *wall;
   u_char check_intersect = 0;

   ccos = cos(catapult->angle * DEG_TO_RAD);
   csin = sin(catapult->angle * DEG_TO_RAD);

   for (i = 0; i < 4; i++) {
      fptptr = &hms.catapult->corner[i];
      *fptptr = catapult->pos;
      fptptr->x += catapult_vec[i].x * ccos - catapult_vec[i].y * csin;
      fptptr->y += catapult_vec[i].x * csin + catapult_vec[i].y * ccos;
      fptptr->z += catapult_vec[i].z;
   }

   for (i = 0; i < hms.num_pieces; i++) {
      wall = hms.pieces + i;
      for (j = 0; j < 4; j++) {
	 d = finddist(&wall->pos[0], &wall->pos[1], &catapult->corner[j]);
	 if (d < 10) return 1;
	 if (d < 1000) check_intersect = 1;
      }
   }
   if (check_intersect) {
      fpoint close[2];
      if (test_intersect(wall->pos, &catapult->corner[0]) == 1) return 1;
      if (test_intersect(wall->pos, &catapult->corner[1]) == 1) return 1;
      if (test_intersect(wall->pos, &catapult->corner[2]) == 1) return 1;
      close[0] = catapult->corner[3];
      close[1] = catapult->corner[0];
      if (test_intersect(wall->pos, &close) == 1) return 1;
   }
   return 0;
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void do_pull(int pull)
{
    if (pull >= 0) {
	hms.catapult->pull += pull;
	if (hms.catapult->pull > 500)
	    hms.catapult->pull = 500;
    }
    else
	hms.catapult->pull = 0;

}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void do_roll(int amount)
{
    fpoint savepos;
    float famt, ccos, csin;

    savepos = hms.catapult->pos;
    famt = (float)amount / 20.0;

    ccos = cos(hms.catapult->angle * DEG_TO_RAD);
    csin = sin(hms.catapult->angle * DEG_TO_RAD);

    hms.catapult->pos.x += famt * ccos;
    hms.catapult->pos.y += famt * csin;

    /* Check for collisions */
    if (catapult_corners() == 1) {
       hms.catapult->pos = savepos;
       catapult_corners();
    }
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void do_rotate(int angle)
{
    float d;
    int i, j, saveangle;
    Castle *wall;

    saveangle = hms.catapult->angle;
    hms.catapult->angle += angle;
    normalize_angle(&hms.catapult->angle);

    /* Check for wall collisions */

    if (catapult_corners() == 1) {
       hms.catapult->angle = saveangle;
       catapult_corners();
    }
    else {
       if (angle != 0) do_pull(0);		// update arm position
    }
}

/*--------------------------------------------------------------*/
/* Load a bomb							*/
/*--------------------------------------------------------------*/

void make_bomb()
{
   Boulder *boulder;
   int i;

   for (i = 0; i < hms.num_boulders; i++) {
      boulder = hms.boulders + i;
      if (boulder->state == 0) break;
   }
   if (i == hms.num_boulders) return;

   if (hms.bombs > 0 && hms.catapult->state == 0) {
      hms.bombs--;
      boulder->hits = -1;	// Indicates a bomb
   }
}

/*--------------------------------------------------------------*/
/* Launch a boulder						*/
/*--------------------------------------------------------------*/

void do_launch(int pull, int angle)
{
   Boulder *boulder;
   int i;

   if (hms.gamestate != ACTIVE) return;

   if (hms.catapult->state == 0 && hms.catapult->pull > 0) {

      if (hms.num_boulders == 0) {
	 // Diagnostic:  this is not supposed to happen.
	 fprintf(stderr, "Error condition:  Launch with no boulder!\n");
	 hms.catapult->state = -1;
	 return;
      }

      for (i = 0; i < hms.num_boulders; i++) {
         boulder = hms.boulders + i;
	 if (boulder->state == 0) break;
      }
      if (i == hms.num_boulders) return;

      if (pull >= 0)
         boulder->zvelocity = 6 * (float)pull;
      else
         boulder->zvelocity = 6 * (float)hms.catapult->pull;

      if (angle >= 0)
	 boulder->angle = angle;
      else
         boulder->angle = hms.catapult->angle;

      boulder->state = 1;
      boulder->fvelocity = 1.0;

      hms.catapult->state = -1;
      hms.catapult->pull = 0;
      do_pull(0);	/* Compute boulder position */
   }
   else if (hms.catapult->state == -1) {
      hms.catapult->state = 0;

      // Generate a new boulder record

      if (hms.num_boulders == 0) {
         hms.num_boulders++;
	 hms.boulders = (Boulder *)Tcl_Alloc(sizeof(Boulder));
         boulder = hms.boulders + hms.num_boulders - 1;
      }
      else {
	 // Recycle used boulder records
	 for (i = 0; i < hms.num_boulders; i++) {
            boulder = hms.boulders + i;
	    if (boulder->state == -2)
	       break;
	 }
	 if (i == hms.num_boulders) {
            hms.num_boulders++;
	    hms.boulders = (Boulder *)Tcl_Realloc(hms.boulders,
			hms.num_boulders * sizeof(Boulder));
            boulder = hms.boulders + hms.num_boulders - 1;
	 }
      }

      boulder->zvelocity = 0;
      boulder->fvelocity = 0.0;
      if (angle >= 0)
	 boulder->angle = angle;
      else
         boulder->angle = hms.catapult->angle;
      boulder->state = 0;
      boulder->hits = 0;
   }
}

/*--------------------------------------------------------------*/
/* More bad guys.						*/
/*--------------------------------------------------------------*/

void generate_new_enemies()
{
    int i, kb, margin, rval;
    int angle, dist, width, height;
    Boulder *boulder, tb;
    Enemy *enemy;
    Target *target;
    fpoint screenpos;

    if (hms.width == 0 || hms.height == 0) return;

    if (hms.num_enemies > 0) Tcl_Free(hms.enemies);
    hms.num_enemies = 2 + hms.level;
    hms.enemies = (Enemy *)Tcl_Alloc(hms.num_enemies * sizeof(Enemy));

    /* Select positions as a random distance and random angle */
    for (i = 0; i < hms.num_enemies; i++) {
	enemy = hms.enemies + i;
        while (1) {
	   angle = (rand() % 2000) - 1000;
	   dist = (rand() % 200) + 300;
	
	   enemy->pos.x = (float)dist * cos((float)angle * DEG_TO_RAD);
	   enemy->pos.y = (float)dist * sin((float)angle * DEG_TO_RAD);
	   enemy->pos.z = 0.0;

	   /* Ensure that this position is not far off-screen */

	   MultiplyByMatrix(&hms.field2screen, &enemy->pos, &screenpos);
	   margin = 10 * hms.level;
	   if (screenpos.x > -margin && screenpos.y > -margin && screenpos.x <
		hms.width + margin && screenpos.y < hms.height + margin)
	      break;
       }
       enemy->state = STOPPED;
       enemy->strategy = ADVANCE;
       enemy->angle = 1800;

       // Some enemies may carry bombs, shields, or extra lives
       rval = rand() % 100;
       if (rval > 95)
	  enemy->flags = LIFE;
       else if (rval > 85)
	  enemy->flags = BOMB;
       else if (rval > 75)
	  enemy->flags = SHIELD;
       else
          enemy->flags = 0;
       set_future_time(&enemy->timeout, 2.0 + ((float)(rand() % 3000) / 1000.0));
    }

    /* The catapult is allowed to remain loaded and in a non-reset	*/
    /* state between levels and through retries.  Look for a boulder	*/
    /* in state zero, and if there is one, keep it in the new list.	*/

    kb = 0;
    for (i = 0; i < hms.num_boulders; i++) {
      boulder = hms.boulders + i;
      if (boulder->state == 0) {
	 tb = *boulder;
	 kb = 1;
      }
    }

    /* Reset the boulders */
    if (hms.num_boulders > 0) {
       Tcl_Free(hms.boulders);
       hms.num_boulders = 0;
       hms.boulders = NULL;
    }

    if (kb == 1) {
       hms.num_boulders = 1;
       hms.boulders = Tcl_Alloc(sizeof(Boulder));
       boulder = hms.boulders;
       *boulder = tb;
    }

    /* Reset all bonus markers */
    if (hms.num_bonuses > 0) {
       Tcl_Free(hms.bonuses);
       hms.num_bonuses = 0;
       hms.bonuses = NULL;
    }

    /* Generate targets */

    if (hms.num_targets > 0) {
       Tcl_Free(hms.targets);
       hms.targets = NULL;
       hms.num_targets = 0;
    }

    hms.num_targets = rand() % 6;

    if (hms.num_targets > 0) {
       hms.targets = (Enemy *)Tcl_Alloc(hms.num_targets * sizeof(Target));

       /* Select positions as a random distance and random angle */
       for (i = 0; i < hms.num_targets; i++) {
	   target = hms.targets + i;
	   while (1) {
	      angle = (rand() % 2000) - 1000;
	      dist = (rand() % 200) + 200;

	      target->pos.x = (float)dist * cos((float)angle * DEG_TO_RAD);
	      target->pos.y = (float)dist * sin((float)angle * DEG_TO_RAD);
	      target->pos.z = 0.0;

	      // Ensure that targets can be seen on the screen
	      MultiplyByMatrix(&hms.field2screen, &target->pos, &screenpos);
	      if (screenpos.x > 0 && screenpos.y > 0 && screenpos.x <
			hms.width && screenpos.y < hms.height)
	         break;
	   }
	   target->state = UNCLAIMED;
	   rval = rand() % 100;
	   if (rval < 40)
	      target->type = SHIELD;
	   else if (rval < 70)
	      target->type = BOMB;
	   else if (rval < 95)
	      target->type = BONUS;
	   else
	      target->type = LIFE;
       }
    }
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void draw_winner()
{
   char pstr[50], *rstr;
   fpoint forig = {0, 0, 1000};
   int width, i;

   SetForeground(WHITE);
   sprintf(pstr, "You Win!");
   width = XTextWidth(hms.fonts[1], pstr, strlen(pstr)); 
   forig.x = (float)(hms.width - width) / 2;
   forig.y = hms.height / 2;
   DrawString(forig, pstr, strlen(pstr), 1);

   SetForeground(YELLOW);
   sprintf(pstr, "Final Score: %d", hms.score);
   width = XTextWidth(hms.fonts[1], pstr, strlen(pstr)); 
   forig.x = (float)(hms.width - width) / 2;
   forig.y += 50;
   DrawString(forig, pstr, strlen(pstr), 1);

   // High score name entry

   if (check_high_score() == 1) {
      Tcl_Eval(helpmeinterp, "wm state .dialog");
      rstr = Tcl_GetStringResult(helpmeinterp);
      if (!strncmp(rstr, "withdrawn", 9)) {
	 Tcl_Eval(helpmeinterp, "wm deiconify .dialog");
      }

      forig.y += 50;

      for (i = 0; i < 10; i++)
	 if (hms.score >= hms.high_scores[i].score)
	    break;

      if (strlen(hms.name) > 0)
         sprintf(pstr, "%s: New high score, rank %d", hms.name, i + 1);
      else
         sprintf(pstr, "New high score, rank %d", i + 1);
      SetForeground(DARK_BLUE);
      width = XTextWidth(hms.fonts[1], pstr, strlen(pstr)); 
      forig.x = (float)(hms.width - width) / 2;
      DrawString(forig, pstr, strlen(pstr), 1);
   }
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void draw_gameover()
{
   char pstr[50];
   fpoint forig = {0, 0, 1000};
   int width;

   SetForeground(WHITE);
   sprintf(pstr, "Game Over");
   width = XTextWidth(hms.fonts[1], pstr, strlen(pstr)); 
   forig.x = (float)(hms.width - width) / 2;
   forig.y = hms.height / 2;
   DrawString(forig, pstr, strlen(pstr), 1);

   SetForeground(YELLOW);
   sprintf(pstr, "Final Score: %d", hms.score);
   width = XTextWidth(hms.fonts[1], pstr, strlen(pstr)); 
   forig.x = (float)(hms.width - width) / 2;
   forig.y = hms.height / 2 + 50,
   DrawString(forig, pstr, strlen(pstr), 1);

   SetForeground(DARK_GREEN);
   sprintf(pstr, "Press Space to restart");
   width = XTextWidth(hms.fonts[1], pstr, strlen(pstr)); 
   forig.x = (float)(hms.width - width) / 2;
   forig.y = hms.height / 2 + 100;
   DrawString(forig, pstr, strlen(pstr), 1);
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void draw_paused()
{
   char pstr[50];
   fpoint forig = {0, 0, 1000};
   int width;

   SetForeground(WHITE);
   sprintf(pstr, "Paused");
   width = XTextWidth(hms.fonts[1], pstr, strlen(pstr)); 
   forig.x = (float)(hms.width - width) / 2;
   forig.y = hms.height / 2 + 50;
   DrawString(forig, pstr, strlen(pstr), 1);
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void draw_interlevel()
{
   char pstr[50];
   fpoint forig = {0, 0, 1000};
   int width;

   SetForeground(WHITE);
   sprintf(pstr, "Level %d", hms.level + 1);
   width = XTextWidth(hms.fonts[1], pstr, strlen(pstr)); 
   forig.x = (float)(hms.width - width) / 2;
   forig.y = hms.height / 2 + 50;
   DrawString(forig, pstr, strlen(pstr), 1);
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void draw_retry()
{
   char pstr[50];
   fpoint forig = {0, 0, 1000};
   int width;

   SetForeground(WHITE);
   sprintf(pstr, "Get ready. . .");
   width = XTextWidth(hms.fonts[1], pstr, strlen(pstr)); 
   forig.x = (float)(hms.width - width) / 2;
   forig.y = hms.height / 2 + 50;
   DrawString(forig, pstr, strlen(pstr), 1);
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void draw_title()
{
   char pstr[50];
   fpoint forig = {0, 0, -100}, screenpos[4];
   int width, i;

   render_init();

   // SetForeground(DARK_BLUE);
   // FillRectangle(forig, hms.width, hms.height);
   
   for (i = 0; i < 4; i++) screenpos[i] = forig;
   screenpos[1].x += hms.width;
   screenpos[2].x += hms.width;
   screenpos[2].y += hms.height;
   screenpos[3].y += hms.height;
   TextureRectangle(screenpos, &hms.blue_map, 1.0);

   forig.z = 100;
   SetForeground(WHITE);
   sprintf(pstr, "Help Me!");
   width = XTextWidth(hms.fonts[1], pstr, strlen(pstr)); 
   forig.x = (float)(hms.width - width) / 2;
   forig.y = hms.height / 2 - 200;
   DrawString(forig, pstr, strlen(pstr), 1);
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void draw_splash()
{
   char pstr[50];
   fpoint forig = {0, 0, 100};
   int width;

   SetForeground(YELLOW);
   sprintf(pstr, "Press Space to start");
   width = XTextWidth(hms.fonts[1], pstr, strlen(pstr)); 
   forig.x = (float)(hms.width - width) / 2;
   forig.y = hms.height / 2;
   DrawString(forig, pstr, strlen(pstr), 1);

   SetForeground(LIGHT_GREEN);
   sprintf(pstr, "Design: William and Michael Edwards");
   width = XTextWidth(hms.fonts[0], pstr, strlen(pstr)); 
   forig.x = (float)(hms.width - width) / 2;
   forig.y = hms.height / 2 + 50;
   DrawString(forig, pstr, strlen(pstr), 0);

   sprintf(pstr, "Programming: Tim Edwards");
   width = XTextWidth(hms.fonts[0], pstr, strlen(pstr)); 
   forig.x = (float)(hms.width - width) / 2;
   forig.y = hms.height / 2 + 80;
   DrawString(forig, pstr, strlen(pstr), 0);
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void draw_scores()
{
   char pstr[50];
   fpoint forig = {0, 0, 100}, fline[2];
   float fx;
   int i, width;

   forig.y = hms.height / 2 - 150;

   SetForeground(YELLOW);
   sprintf(pstr, "High Scores:");
   width = XTextWidth(hms.fonts[1], pstr, strlen(pstr)); 
   forig.x = (float)(hms.width - width) / 2;
   DrawString(forig, pstr, strlen(pstr), 1);

   SetForeground(LIGHT_BLUE);
   fline[0].y = fline[1].y = forig.y + 40;
   fline[0].z = fline[1].z = 100;
   fline[0].x = 50;
   fline[1].x = hms.width - 50;
   DrawLine(&fline[0], &fline[1]);
   SetForeground(LIGHT_BLUE);
   fline[0].y = fline[1].y = forig.y + 80;
   DrawLine(&fline[0], &fline[1]);

   forig.y += 60;
   forig.x = hms.width / 2 - 250;
   fx = forig.x;

   SetForeground(LIGHT_BLUE);
   forig.x = fx;
   sprintf(pstr, "Name");
   DrawString(forig, pstr, strlen(pstr), 0);
   forig.x = fx + 150;
   sprintf(pstr, "Date");
   DrawString(forig, pstr, strlen(pstr), 0);
   forig.x = fx + 250;
   sprintf(pstr, "Level");
   DrawString(forig, pstr, strlen(pstr), 0);
   forig.x = fx + 350;
   sprintf(pstr, "Score");
   DrawString(forig, pstr, strlen(pstr), 0);

   forig.y += 60;
   SetForeground(LIGHT_GREEN);
   for (i = 0; i < 10; i++) {
      if (hms.high_scores[i].score == 0) break;
      forig.x = fx;
      strcpy(pstr, hms.high_scores[i].name);
      DrawString(forig, pstr, strlen(pstr), 0);

      forig.x = fx + 150;
      strcpy(pstr, hms.high_scores[i].date);
      DrawString(forig, pstr, strlen(pstr), 0);

      forig.x = fx + 250;
      sprintf(pstr, "%d", hms.high_scores[i].level);
      DrawString(forig, pstr, strlen(pstr), 0);

      forig.x = fx + 350;
      sprintf(pstr, "%d", hms.high_scores[i].score);
      DrawString(forig, pstr, strlen(pstr), 0);

      forig.y += 30;
   }
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void draw_help()
{
   char pstr[50];
   fpoint forig = {0, 0, 100}, fline[2];
   float fx;
   int i, width;

   forig.y = hms.height / 2 - 150;

   SetForeground(YELLOW);
   sprintf(pstr, "Instructions:");
   width = XTextWidth(hms.fonts[1], pstr, strlen(pstr)); 
   forig.x = (float)(hms.width - width) / 2;
   DrawString(forig, pstr, strlen(pstr), 1);

   SetForeground(LIGHT_BLUE);
   fline[0].y = fline[1].y = forig.y + 10;
   fline[0].z = fline[1].z = 100;
   fline[0].x = 100;
   fline[1].x = hms.width - 100;
   DrawLine(&fline[0], &fline[1]);

   forig.x = 200;
   forig.y += 50;

   sprintf(pstr, "Hit the soldiers with catapult boulders before");
   DrawString(forig, pstr, strlen(pstr), 0);
   forig.y += 30;
   sprintf(pstr, "they reach you.  Strike target spots to get");
   DrawString(forig, pstr, strlen(pstr), 0);
   forig.y += 30;
   sprintf(pstr, "materials for walls and bombs, or bonus lives");
   DrawString(forig, pstr, strlen(pstr), 0);
   forig.y += 30;
   sprintf(pstr, "and points.  Build a castle to protect yourself");
   DrawString(forig, pstr, strlen(pstr), 0);
   forig.y += 30;
   sprintf(pstr, "and win the game.");
   DrawString(forig, pstr, strlen(pstr), 0);
   forig.y += 60;

   sprintf(pstr, "Catapult forward/backward: Direction keys, scroll wheel");
   DrawString(forig, pstr, strlen(pstr), 0);
   forig.y += 30;
   sprintf(pstr, "Catapult rotate:  Direction keys, mouse right button");
   DrawString(forig, pstr, strlen(pstr), 0);
   forig.y += 30;
   sprintf(pstr, "Load or fire:  mouse left button");
   DrawString(forig, pstr, strlen(pstr), 0);
   forig.y += 30;
   sprintf(pstr, "Pull catapult arm:  mouse middle button");
   DrawString(forig, pstr, strlen(pstr), 0);
   forig.y += 30;
   sprintf(pstr, "Turn boulder into bomb:  B key");
   DrawString(forig, pstr, strlen(pstr), 0);
   forig.y += 30;
   sprintf(pstr, "Place or build castle wall:  C key");
   DrawString(forig, pstr, strlen(pstr), 0);
   forig.y += 30;
   sprintf(pstr, "Pause game:  P key");
   DrawString(forig, pstr, strlen(pstr), 0);
   forig.y += 30;
   sprintf(pstr, "Quit game:  Q key");
   DrawString(forig, pstr, strlen(pstr), 0);
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void add_bonus(int value, fpoint pos)
{
   Bonus *bonus;
   fpoint screenpos;

   hms.num_bonuses++;
   if (hms.num_bonuses == 1)
      hms.bonuses = Tcl_Alloc(sizeof(Bonus));
   else
      hms.bonuses = Tcl_Realloc(hms.bonuses, hms.num_bonuses * sizeof(Bonus));

   bonus = hms.bonuses + hms.num_bonuses - 1;
   
   bonus->value = value;
   set_future_time(&bonus->timeout, 2.0);

   MultiplyByMatrix(&hms.field2screen, &pos, &screenpos);
   bonus->pos.x = screenpos.x;
   bonus->pos.y = screenpos.y - 10; 	// Raise slightly 
   bonus->pos.z = 1000;			// Make sure it's on top.
}

/*--------------------------------------------------------------*/
/* The position of the last wall piece tracks the pointer	*/
/* position for as long as it remains unbuilt.			*/
/* Return 0 if nothing is in the way, 1 if the wall can't be	*/
/* placed due to an obstruction.				*/
/*--------------------------------------------------------------*/

u_char set_wall_pos()
{
    fpoint wall_vec[2] = {
	{50.0,  0.0,  0.0},
	{ 0.0,  0.0,  0.0},
    };

    int i, nullint, x, y, quadrant;
    Window nullwin;
    u_int  nullui;
    fpoint fp, refpt, savept[2], close[2];
    Castle *wall, *chkwall;
    Enemy *enemy;
    u_char obstruct = 0;
    
    if (hms.num_pieces == 0) return;

    for (i = 0; i < hms.num_pieces; i++) {
       wall = hms.pieces + i;
       if (wall->type == UNPLACED) break;
    }
    if (i == hms.num_pieces) return;	// No unbuilt wall

    savept[0] = wall->pos[0];
    savept[1] = wall->pos[1];

    XQueryPointer(dpy, Tk_WindowId(hms.area), &nullwin, &nullwin,
		&nullint, &nullint, &x, &y, &nullui);
    get_field_position(x, y, &fp);

    fp.x += 10000;	// This is easier than handling negative numbers
    fp.y += 10000;
    x = (int)(fp.x) % 50;
    y = (int)(fp.y) % 50;

    if (x < y)
        quadrant = (x < (50 - y)) ? 1 : 4;
    else
        quadrant = (x < (50 - y)) ? 2 : 3;

    x = (int)fp.x / 50.0;
    y = (int)fp.y / 50.0;
    refpt.x = (float)(x * 50) - 10000;
    refpt.y = (float)(y * 50) - 10000;

    if (quadrant == 3) refpt.x += 50.0;
    if (quadrant == 4) refpt.y += 50.0;

    if (quadrant == 2 || quadrant == 4) {
       for (i = 0; i < 2; i++) {
          wall->pos[i].x = wall_vec[i].x + refpt.x;
          wall->pos[i].y = refpt.y;
          wall->pos[i].z = wall_vec[i].z;
       }
       wall->align = XALIGNED;
    }
    else {
       for (i = 0; i < 2; i++) {
          wall->pos[i].x = refpt.x;
          wall->pos[i].y = wall_vec[i].x + refpt.y;
          wall->pos[i].z = wall_vec[i].z;
       }
       wall->align = YALIGNED;
    }

    /* Check for obstructions---catapult, boulders, enemies, or */
    /* other walls						*/

    for (i = 0; i < hms.num_pieces; i++) {
       chkwall = hms.pieces + i;
       if (chkwall->type != BUILT) continue;
       if (wall->pos[0].x == chkwall->pos[0].x &&
		wall->pos[0].y == chkwall->pos[0].y &&
		wall->align == chkwall->align) {
	  obstruct = 1;
	  break;
       }
    }

    for (i = 0; i < hms.num_enemies; i++) {
       enemy = hms.enemies + i;
       if (finddist(&wall->pos[0], &wall->pos[1], &enemy->pos) < 80) {
	  obstruct = 1;
	  break;
       }
    }

    if (test_intersect(wall->pos, &hms.catapult->corner[0]) == 1)
       obstruct = 1;
    else if (test_intersect(wall->pos, &hms.catapult->corner[1]) == 1) 
       obstruct = 1;
    else if (test_intersect(wall->pos, &hms.catapult->corner[2]) == 1) 
       obstruct = 1;
    else {
       close[0] = hms.catapult->corner[3];
       close[1] = hms.catapult->corner[0];
       if (test_intersect(wall->pos, &close) == 1) obstruct = 1;
    }

    if (obstruct) {
       wall->pos[0] = savept[0];
       wall->pos[1] = savept[1];
    }
    return obstruct;
}
	
/*--------------------------------------------------------------*/
/* Compute new positions for everything				*/
/*--------------------------------------------------------------*/

void recalculate()
{
   int i, j, bonus, lmax, diff, angs, damage, rval, nt, ne, tidx;
   float adv, bdist, ang, tdist, tmin, dist;
   fpoint fdist, newpos, tmppos;
   Boulder *boulder;
   Enemy *enemy, *other;
   Target *target;
   Castle *wall;

   // Get time of this step

   gettimeofday(&hms.now, NULL);
   if (hms.paused > 0) return;

   // for non-active states. . .

   switch (hms.gamestate) {
      case SPLASH:
      case SCORES:
      case INSTRUCTIONS:
	 if (compare_time(&hms.timeout)) {
	    hms.gamestate = (hms.gamestate == INSTRUCTIONS) ? SPLASH :
		hms.gamestate + 1;
	    set_future_time(&hms.timeout, 5.0);
	 }
      case GAMEOVER:
      case WINNER:
	 return;
      case RETRY:
         if (compare_time(&hms.timeout)) {
	    hms.gamestate = ACTIVE;
	    generate_new_enemies();
	    return;
	 }
	 break;
      case INTERLEVEL:
         if (compare_time(&hms.timeout)) {
	    hms.gamestate = ACTIVE;
	    hms.level++;
	    generate_new_enemies();
	    return;
	 }
	 break;
   }

   // Once every 5 seconds, check number of targeted targets and
   // number of enemies, and select enemies for grabbing nearby targets

   if (compare_time(&hms.timeout)) {
      nt = 0;
      for (i = 0; i < hms.num_targets; i++) {
         target = hms.targets + i;
         if (target->state == UNCLAIMED) nt++;
      }
      ne = 0;
      for (i = 0; i < hms.num_enemies; i++) {
         enemy = hms.enemies + i;
         if (enemy->state != DEAD) ne++;
      }
      if ((nt > 0) && (ne >= 3)) {
	 for (i = 0; i < hms.num_targets; i++) {
            target = hms.targets + i;
            if (target->state == UNCLAIMED) {
	       tmin = 100000.0;
	       for (j = 0; j < hms.num_enemies; j++) {
		  enemy = hms.enemies + j;
		  if ((enemy->state == DEAD) || (enemy->flags != 0) ||
				(enemy->strategy >= GET_TARGET))
		     continue;
		  tdist = fsqlinelen(&enemy->pos, &target->pos);
		  if (tdist < tmin) {
		     tmin = tdist;
		     tidx = j;
		  }
	       }
	       if (tmin < 100000.0) {
		  enemy = hms.enemies + tidx;
		  enemy->strategy = GET_TARGET + i;
		  target->state = TARGETED;
		  ne--;
		  if (ne < 3) break;
	       }
	    }
	 }
      }
      set_future_time(&hms.timeout, 5.0);
   }

   // active state:  enemies advance, etc.

   lmax = (hms.level > 12) ? 12 : hms.level;
   adv = (float)lmax * 0.01;

   for (i = 0; i < hms.num_enemies; i++) {
      if (hms.gamestate != ACTIVE) break;
      enemy = hms.enemies + i;
      if (enemy->state == DEAD) {
	 if ((enemy->flags != 0) && compare_time(&enemy->timeout)) {
	    enemy->flags = 0;
	    enemy->state = STOPPED;
	    enemy->strategy = ADVANCE;
            set_future_time(&enemy->timeout, 1.0);
	 }
	 else
	    continue;
      }
      newpos = enemy->pos;

      // If strategy has expired, find a new strategy
      // "GET_TARGET" strategy does not have a timeout.

      if ((enemy->strategy < GET_TARGET) && compare_time(&enemy->timeout)) {
	 if (enemy->strategy == RETREAT)
	    enemy->strategy = BACK_OFF;
	 else {
	    rval = rand() % 100;
	    if (rval < 50)
	       enemy->strategy = ADVANCE;
	    else if (rval < 75)
	       enemy->strategy = ARC_LEFT;
	    else
	       enemy->strategy = ARC_RIGHT;
	 }
         set_future_time(&enemy->timeout, 2.0 + ((float)(rand() % 3000) / 1000.0));
      }

      if (enemy->strategy != SCATTER) {
         if (rand() % 50 == 1) {
	    if (enemy->state == MOVING)
	       enemy->state = STOPPED;
	    else if (enemy->state == STOPPED)
	       enemy->state = MOVING;
         }
         if (enemy->state == STOPPED) continue;
      }

      switch (enemy->strategy) {
	 case ADVANCE:
	 case ARC_LEFT:
	 case ARC_RIGHT:
	 case SCATTER:
	 case BACK_OFF:
	    /* Compute angle between enemy and the catapult */
	    ang = atan2(hms.catapult->pos.y - enemy->pos.y,
			enemy->pos.x - hms.catapult->pos.x);
	    ang *= RAD_TO_DEG;
	    diff = 1800 - (int)ang;
	    normalize_angle(&diff);
	    switch (enemy->strategy) {
	       case ARC_LEFT:
	          diff += 600;
		  break;
	       case ARC_RIGHT:
	          diff -= 600;
		  break;
	       case BACK_OFF:
		  diff -= 1800;
		  break;
	       case SCATTER:
		  angs = 1800 + diff;
		  normalize_angle(&angs);
		  if (angs > hms.catapult->angle)
		     diff -= 900;
		  else
		     diff += 900;
		  break;
	    }
	    normalize_angle(&diff);
	    diff -= enemy->angle;
	    normalize_angle(&diff);
	    if (enemy->strategy == SCATTER)
	       enemy->angle += (diff > 0) ? 5 : -5;
	    else
	       enemy->angle += (diff > 0) ? 1 : -1;
	    normalize_angle(&enemy->angle);
	    break;

	 case RETREAT:		// No change in angle during retreat
	    break;

	 default:
	    target = hms.targets + (enemy->strategy - GET_TARGET);
	    if (target->state == GONE) {
	       enemy->strategy = ADVANCE;
	       set_future_time(&enemy->timeout, 3.0);
	    }

	    /* Compute angle between enemy and the target */
	    ang = atan2(target->pos.y - enemy->pos.y,
			enemy->pos.x - target->pos.x);
	    ang *= RAD_TO_DEG;
	    diff = 1800 - (int)ang;
	    normalize_angle(&diff);
	    diff -= enemy->angle;
	    normalize_angle(&diff);
	    enemy->angle += (diff > 0) ? 1 : -1;
	    normalize_angle(&enemy->angle);
	    break;
      }

      // Advance position in the direction indicated by angle
	
      if (enemy->strategy != BACK_OFF) {
         dist = adv * cos(enemy->angle * DEG_TO_RAD);
         if (enemy->strategy == RETREAT) dist = -dist;
         newpos.x += dist;
         dist = adv * sin(enemy->angle * DEG_TO_RAD);
         if (enemy->strategy == RETREAT) dist = -dist;
         newpos.y += dist;
      }

      // Check if position crossed a wall boundary.  If so, the
      // enemy can't go there.

      for (j = 0; j < hms.num_pieces; j++) {
	 wall = hms.pieces + j;
	 if (wall->type != BUILT) continue;
	 if (finddist(&wall->pos[0], &wall->pos[1], &newpos) < 70) {
	    if (enemy->flags == BOMB) {
	       enemy->flags = 0;
	       wall->type = DESTROYED;
	    }
	    else {
	       newpos = enemy->pos;
	       enemy->strategy = RETREAT;
	       set_future_time(&enemy->timeout,
			1.0 + ((float)(rand() % 3000) / 1000.0));
	    }
	    break;
	 }
      }

      // Check if position crossed another enemy position.  If so, the
      // enemy can't go there.

      for (j = 0; j < hms.num_enemies; j++) {
	 if (j == i) continue;
	 other = hms.enemies + j;
	 if (other->state == DEAD) continue;
	 if (fsqlinelen(&newpos, &other->pos) < 260.0) {
	    newpos = enemy->pos;
	    if (enemy->strategy == RETREAT) {
	       enemy->strategy = ADVANCE;
	       enemy->state = STOPPED;
	    }
	    else {
	       enemy->strategy = RETREAT;
	       set_future_time(&enemy->timeout,
			1.0 + ((float)(rand() % 1000) / 1000.0));
	    }
	    break;
	 }
      }

      // Check if position crossed a stationary boulder position.  If so, the
      // enemy can't go there.

      for (j = 0; j < hms.num_boulders; j++) {
	 boulder = hms.boulders + j;
	 if (boulder->state == -1) {
	    if (fsqlinelen(&newpos, &boulder->pos) < 330.0) {
	       newpos = enemy->pos;
	       if (enemy->strategy == RETREAT) {
	          enemy->strategy = ADVANCE;
	          enemy->state = STOPPED;
	       }
	       else {
	          enemy->strategy = RETREAT;
	          set_future_time(&enemy->timeout,
			1.0 + ((float)(rand() % 1000) / 1000.0));
	       }
	       break;
	   }
	 }
      }
      enemy->pos = newpos;

      /* Use a simple calculation to avoid unnecessary computation */

      if (fsqlinelen(&enemy->pos, &hms.catapult->pos) < 4000.0) {
         u_char collide = 0;
	 fpoint *ccorner = hms.catapult->corner;
	 if (finddist(ccorner, ccorner + 1, &enemy->pos) < 10.0)
	    collide = 1;
	 else if (finddist(ccorner + 1, ccorner + 2, &enemy->pos) < 10.0)
	    collide = 1;
	 else if (finddist(ccorner + 2, ccorner + 3, &enemy->pos) < 10.0)
	    collide = 1;
	 else if (finddist(ccorner + 3, ccorner, &enemy->pos) < 10.0)
	    collide = 1;

         if (collide == 1) {
	    hms.lives--;
	    if (hms.lives <= 0) {
	       // game over
	       hms.gamestate = GAMEOVER;
	    }
	    else {
	       hms.gamestate = RETRY;
	       set_future_time(&hms.timeout, 3.0);
	    }
         }
      }

      // For any enemy whose strategy is to grab a target, see if it has
      // been reached.

      if (enemy->strategy >= GET_TARGET) {
	 target = hms.targets + (enemy->strategy - GET_TARGET);
	 if (fsqlinelen(&enemy->pos, &target->pos) < 100.0) {
	    target->state = GONE;
	    if (target->type == BONUS) {
	       bonus = 500 * hms.level;
	       hms.score -= bonus;
	       if (hms.score < 0) hms.score = 0;
	    }
	    else
	       enemy->flags = target->type;
	 }
      }
   }

   if (hms.active_pull) {
      do_pull(1);
   }
   if (hms.active_rotate != 0) {
      do_rotate(hms.active_rotate);
   }
   if (hms.active_roll != 0) {
      do_roll(hms.active_roll);
   }

   for (i = 0; i < hms.num_boulders; i++) {
      boulder = hms.boulders + i;
      if (boulder->state > 0) {
         float vspeed;

         tmppos.x = boulder->pos.x + boulder->fvelocity
			* cos(boulder->angle * DEG_TO_RAD);
         tmppos.y = boulder->pos.y + boulder->fvelocity
			* sin(boulder->angle * DEG_TO_RAD);
         vspeed = (float)boulder->zvelocity / 500;
         boulder->pos.z += vspeed;
         boulder->zvelocity -= 10;

	 // Boulders can collide with castle walls
	 if (boulder->pos.z < 50) {
	    for (j = 0; j < hms.num_pieces; j++) {
	       wall = hms.pieces + j;
	       if (wall->type != BUILT) continue;
	       if (finddist(&wall->pos[0], &wall->pos[1], &tmppos) < 100) {
		  if ((boulder->pos.z < 40) || (boulder->zvelocity > 0)) {
		     if (wall->align == XALIGNED)
		        boulder->angle = -boulder->angle;
		     else
		        boulder->angle = 1800 - boulder->angle;
		     normalize_angle(&boulder->angle);
                     tmppos.x = boulder->pos.x + cos(boulder->angle * DEG_TO_RAD);
                     tmppos.y = boulder->pos.y + sin(boulder->angle * DEG_TO_RAD);
		     // Walls take damage!
		     damage = (int)(boulder->fvelocity * 60.0);
		     if (wall->strength <= damage)
			wall->type = DESTROYED;
		     else
		        wall->strength -= (u_char)damage;
		  } else {
	             boulder->state++;
	             boulder->zvelocity = -boulder->zvelocity;
		     if (wall->strength <= 10)
		        wall->type = DESTROYED;
		     else
		        wall->strength -= 10;
		  }
	          boulder->fvelocity = boulder->fvelocity * 0.5;
		  break;
	       }
	    }
	 }
         boulder->pos.x = tmppos.x;
         boulder->pos.y = tmppos.y;

         // Bounce!
         if (boulder->pos.z <= 10.0) {
	    if (boulder->hits < 0) {
	       /* Bombs explode when they hit */
	       boulder->zvelocity = 0;
	       boulder->fvelocity = 0;
	       boulder->state = -1;
	       set_future_time(&boulder->timeout, 1.0);
	    }
	    else {
	       boulder->state++;
	       boulder->zvelocity = -boulder->zvelocity * 0.5;
	       boulder->fvelocity = boulder->fvelocity * 0.9;
               boulder->pos.z = 10.0;
	    }
         }

	 if (hms.gamestate == ACTIVE) {
            for (j = 0; j < hms.num_targets; j++) {
	       target = hms.targets + j;
	       if (target->state == GONE) continue;
	       fdist.x = fabs(target->pos.x - boulder->pos.x);
	       fdist.y = fabs(target->pos.y - boulder->pos.y);
	       fdist.z = boulder->pos.z - 10;
	       if (fdist.x < 10 && fdist.y < 10 && fdist.z < 5) {
		  switch (target->type) {
		     case BONUS:
		        bonus = 500 * hms.level;
		        hms.score += bonus;
	 	        add_bonus(bonus, target->pos);		
			target->state = GONE;
			break;
		     case LIFE:
		        hms.lives++;
			target->state = GONE;
			break;
		     case BOMB:
		        hms.bombs++;
			target->state = GONE;
			if (check_victory() == 1) {
			   hms.gamestate = WINNER;
			}
			break;
		     case SHIELD:
		        hms.shields++;
			target->state = GONE;
			break;
		  }
	       }
	    }

            tmppos.x = boulder->pos.x + 1000 * cos(boulder->angle * DEG_TO_RAD);
            tmppos.y = boulder->pos.y + 1000 * sin(boulder->angle * DEG_TO_RAD);
	    tmppos.z = 0;

            for (j = 0; j < hms.num_enemies; j++) {
	       enemy = hms.enemies + j;
	       if (enemy->state == DEAD) continue;
	       fdist.x = fabs(enemy->pos.x - boulder->pos.x);
	       fdist.y = fabs(enemy->pos.y - boulder->pos.y);
	       fdist.z = fabs(enemy->pos.z - boulder->pos.z);
	       if (boulder->hits >= 0 && fdist.x < 10
				&& fdist.y < 10 && fdist.z < 30) {
	          enemy->state = DEAD;
		  if (enemy->flags == LIFE || enemy->flags == SHIELD)
	             set_future_time(&enemy->timeout, 1.0);
		  else {
	             hms.score += 100 * hms.level;
		     enemy->flags = 0;
		     boulder->hits++;
		  }

	          // Add a direct-hit bonus (hit with no bounces)
		  if (enemy->flags == 0) {
	             if (boulder->state == 1) {
		        bonus = 75 * hms.level;
		        hms.score += bonus;
	 	        add_bonus(bonus, boulder->pos);		
		     }

	             // Add a multiple-hit bonus (hit with no bounces)
	             if (boulder->hits > 1) {
		        bonus = 25 * hms.level * (boulder->hits - 1);
		        hms.score += bonus;
	 	        add_bonus(bonus, boulder->pos);		
		     }
		  }
	       }

	       // If enemy is near boulder's line-of-flight, run!
	       else if (enemy->strategy != SCATTER) {
		  if (boulder->zvelocity < 0) {
		     if (finddist(&boulder->pos, &tmppos, &enemy->pos) < 1000) {
			if ((enemy->strategy < GET_TARGET) || ((rand() % 1000) < 1)) {
		           enemy->strategy = SCATTER;
	                   set_future_time(&enemy->timeout, 1.0);
			}
		     }
		  }
	       }
	    }
         }
  
         if ((boulder->hits >= 0) && (boulder->fvelocity <= 0.001)) {
	    boulder->fvelocity = 0;
	    boulder->state = -1;
	    set_future_time(&boulder->timeout, 10.0);
         }
      }
      else if (boulder->state == -1) {
	 if (compare_time(&boulder->timeout)) {
	    boulder->state = -2;
	 }

	 /* Velocity tracks the exploding bomb circle */
	 if (boulder->hits < 0) {
	    boulder->fvelocity = boulder->fvelocity + 0.08;
            for (j = 0; j < hms.num_enemies; j++) {
	       if (hms.gamestate != ACTIVE) break;
	       enemy = hms.enemies + j;
	       if (enemy->state == DEAD) continue;
	       fdist.x = enemy->pos.x - boulder->pos.x;
	       fdist.y = enemy->pos.y - boulder->pos.y;
	       fdist.x *= fdist.x;
	       fdist.y *= fdist.y;
	       bdist = (boulder->fvelocity * 8);
	       bdist *= bdist;

	       /* Blown away by the blast zone (but no bonuses, and less score) */
	       if ((fdist.x + fdist.y) < bdist) {
	          hms.score += 25 * hms.level;
		  if (enemy->flags != SHIELD) {
		    enemy->state = DEAD;
		    if (enemy->flags == LIFE)
	               set_future_time(&enemy->timeout, 1.0);
		    else
		       enemy->flags = 0;
		  }
	       }
	    }
         }
      }
   }


   // See if there are any pieces to place, and move them.
   if (hms.gamestate == ACTIVE) set_wall_pos();

   // Check if we have cleared the level

   if (hms.gamestate == ACTIVE) {
      for (j = 0; j < hms.num_enemies; j++) {
         enemy = hms.enemies + j;
         if ((enemy->state != DEAD) || (enemy->flags != 0)) break;
      }
      if (j == hms.num_enemies) {
         hms.gamestate = INTERLEVEL;
	 set_future_time(&hms.timeout, 3.0);
      }
   }
}

/*--------------------------------------------------------------*/
/* Drawing routines for each of the separate components that	*/
/* needs to be drawn back-to-front.  This does not include	*/
/* background objects like targets and dead enemy markers, or	*/
/* foreground objects like bonus markers and information text.	*/
/*--------------------------------------------------------------*/

void draw_catapult(Catapult *catapult, fpoint *circ_vec)
{
   fpoint catapult_vec[4] = {
	{ 30.0,  20.0,  0.0},
	{ 30.0, -20.0,  0.0},
	{-30.0, -20.0,  0.0},
	{-30.0,  20.0,  0.0}
   };

   fpoint fieldpos, screenpos[13], bpos;
   int i, j, k;
   float csin, ccos, size;

   ccos = cos(catapult->angle * DEG_TO_RAD);
   csin = sin(catapult->angle * DEG_TO_RAD);

   // Catapult wheels (turn circ_vec upright by swapping Y and Z)

   for (j = 0; j < 4; j++) {
      bpos = catapult->pos;
      bpos.x += catapult_vec[j].x * 0.7 * ccos - catapult_vec[j].y * 1.05 * csin;
      bpos.y += catapult_vec[j].x * 0.7 * csin + catapult_vec[j].y * 1.05 * ccos;
      bpos.z += catapult_vec[j].z;
      for (i = 0; i < 12; i++) {
         fieldpos.x = bpos.x + circ_vec[i].x * ccos - circ_vec[i].z * csin;
         fieldpos.y = bpos.y + circ_vec[i].x * csin + circ_vec[i].z * ccos;
         fieldpos.z = bpos.z + circ_vec[i].y;
         MultiplyByMatrix(&hms.field2screen, &fieldpos, &screenpos[i]);
      }
      SetForeground(BROWN);
      FillPolygon(screenpos, 12);
      for (i = 0; i < 12; i++) {
	 screenpos[i].z += 1.0;
      }
      SetForeground(BLACK);
      DrawLines(screenpos, 12, 1);
   }

   // Draw the catapult base and its shadow
   for (k = 0; k < 2; k++) {
      for (i = 0; i < 4; i++) {
	 fieldpos = catapult->corner[i];
	 fieldpos.z -= (float)(BASELINE * (1 - k));
         MultiplyByMatrix(&hms.field2screen, &fieldpos, &screenpos[i]);
      }
      if (k == 0) {
         SetForeground(DARK_GREEN);
         FillPolygon(screenpos, 4);
      }
      else {
         // SetForeground(TAN);
	 TextureRectangle(screenpos, &hms.wood_map, 1.0);
      }
   }

   // Draw the catapult arm

   bpos = catapult->pos;
   MultiplyByMatrix(&hms.field2screen, &bpos, &screenpos[0]);

   calc_arm_pos(&bpos);
   MultiplyByMatrix(&hms.field2screen, &bpos, &screenpos[1]);
   SetForeground(BROWN);
   DrawLines(screenpos, 2, 0);
}

/*--------------------------------------------------------------*/
/* Draw an enemy soldier					*/
/*--------------------------------------------------------------*/

void draw_enemy(Enemy *enemy, fpoint *circ_vec)
{
   fpoint enemy_vec[6] = {
	{ 0.0,   8.0,  0.0},
	{ 0.0,   8.0, 20.0},
	{ 0.0,   0.0, 25.0},
	{ 0.0,  -8.0, 20.0},
	{ 0.0,  -8.0,  0.0},
	{ 0.0,   0.0, 30.0}
   };
   fpoint enemy_shadow[4] = {
	{ 0.0,   8.0, 0.0},
	{ 5.0,   0.0, 0.0},
	{ 0.0,  -8.0, 0.0},
	{-5.0,   0.0, 0.0}
   };
   fpoint enemy_shield[5] = {
	{6.0,   0.0,  2.0},
	{6.0,   6.0,  5.0},
	{6.0,   6.0, 15.0},
	{6.0,  -6.0, 15.0},
	{6.0,  -6.0,  5.0}
   };

   fpoint fieldpos, screenpos[5];
   float csin, ccos;
   int j, k;

   if ((enemy->state == DEAD) && (enemy->flags == 0)) return;

   ccos = cos(enemy->angle * DEG_TO_RAD);
   csin = sin(enemy->angle * DEG_TO_RAD);

   for (j = 0; j < 4; j++) {
      fieldpos = enemy->pos;
      fieldpos.x += enemy_shadow[j].x * ccos - enemy_shadow[j].y * csin;
      fieldpos.y += enemy_shadow[j].x * csin + enemy_shadow[j].y * ccos;
      MultiplyByMatrix(&hms.field2screen, &fieldpos, &screenpos[j]);
   }
   SetForeground(DARK_GREEN);
   FillPolygon(screenpos, 4);

   for (j = 0; j < 5; j++) {
      fieldpos = enemy->pos;
      fieldpos.x += enemy_vec[j].x * ccos - enemy_vec[j].y * csin;
      fieldpos.y += enemy_vec[j].x * csin + enemy_vec[j].y * ccos;
      fieldpos.z += enemy_vec[j].z;
      MultiplyByMatrix(&hms.field2screen, &fieldpos, &screenpos[j]);
   }
   if (enemy->flags == LIFE) {
      if (enemy->state == DEAD)
         SetForeground(WHITE);
      else
         SetForeground(YELLOW);
   }
   else
      SetForeground(RED);
   FillPolygon(screenpos, 5);
   fieldpos.x = enemy->pos.x + enemy_vec[5].x;
   fieldpos.y = enemy->pos.y + enemy_vec[5].y;
   fieldpos.z = enemy->pos.z + enemy_vec[5].z;
   MultiplyByMatrix(&hms.field2screen, &fieldpos, &screenpos[0]);
   SetForeground(TAN);
   FillArc(screenpos[0], 10, circ_vec);

   if (enemy->flags == BOMB) {
      fieldpos = enemy->pos;
      fieldpos.x += 10.0 * ccos;
      fieldpos.y += 10.0 * csin;
      fieldpos.z += 10.0;
      MultiplyByMatrix(&hms.field2screen, &fieldpos, &screenpos[0]);
      SetForeground(BLACK);
      FillArc(screenpos[0], 13, circ_vec);
   }
   else if (enemy->flags == SHIELD) {
      for (j = 0; j < 5; j++) {
         fieldpos = enemy->pos;
         fieldpos.x += enemy_shield[j].x * ccos - enemy_shield[j].y * csin;
         fieldpos.y += enemy_shield[j].x * csin + enemy_shield[j].y * ccos;
         fieldpos.z += enemy_shield[j].z;
         MultiplyByMatrix(&hms.field2screen, &fieldpos, &screenpos[j]);
      }
      SetForeground(LIGHT_BLUE);
      FillPolygon(screenpos, 5);
   }
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void draw_boulder(Boulder *boulder, fpoint *circ_vec)
{
   fpoint fieldpos, screenpos[12];
   float size;
   int i;

   // Draw the boulder's shadow.

   SetForeground(DARK_GREEN);

   if ((boulder->state != -2 && boulder->state != 0) &&
		(!(boulder->state == -1 && boulder->hits < 0))) {
      size = 1.2 - (boulder->pos.z / 300);
      if (size > 0) {
         for (i = 0; i < 12; i++) {
               fieldpos.x = boulder->pos.x + size * circ_vec[i].x;
               fieldpos.y = boulder->pos.y + size * circ_vec[i].y;
               fieldpos.z = 0;
               MultiplyByMatrix(&hms.field2screen, &fieldpos, &screenpos[i]);
         }
         FillPolygon(screenpos, 12);
      }
   }

   // Draw the boulder

   if (boulder->state == 0)
      calc_arm_pos(&boulder->pos);

   if ((boulder->state == -1) && (boulder->hits < 0)) {
      size = boulder->fvelocity;
      for (i = 0; i < 12; i++) {
         fieldpos.x = boulder->pos.x + size * circ_vec[i].x;
         fieldpos.y = boulder->pos.y + size * circ_vec[i].y;
         fieldpos.z = 0;
         MultiplyByMatrix(&hms.field2screen, &fieldpos, &screenpos[i]);
      }
      SetForeground(ORANGE);
      DrawLines(screenpos, 12, 1);
   }
   else if (boulder->state != -2) {
      MultiplyByMatrix(&hms.field2screen, &boulder->pos, &screenpos[0]);
      if (boulder->hits < 0)
	 SetForeground(BLACK);
      else
         SetForeground(BROWN);
      screenpos[0].z += 10;	// Bring forward slightly
      FillArc(screenpos[0], 20, circ_vec);
   }
}

/*--------------------------------------------------------------*/
/* Draw a castle wall						*/
/*--------------------------------------------------------------*/

void draw_castle(Castle *wall)
{
   TextureMap *mapptr;
   fpoint fieldpos[4], screenpos[4];
   int i, k;

   for (i = 0; i < 2; i++) {
      fieldpos[i].x = wall->pos[i].x;
      fieldpos[i].y = wall->pos[i].y;
      fieldpos[i].z = wall->pos[i].z;
      MultiplyByMatrix(&hms.field2screen, &fieldpos[i], &screenpos[i]);
   }

   if (wall->type == UNPLACED) {
      SetForeground(DARK_GRAY);
      DrawLines(screenpos, 2, 0);
   }
   else if (wall->type != DESTROYED) {
      if (wall->align == XALIGNED)
	 mapptr = &hms.stone_map;
         // SetForeground(GRAY);
      else
	 mapptr = &hms.stone2_map;
         // SetForeground(DARK_GRAY);
      for (i = 0; i < 2; i++) {
         fieldpos[2 + i].x = wall->pos[i].x;
         fieldpos[2 + i].y = wall->pos[i].y;
         fieldpos[2 + i].z = wall->pos[i].z + 40;
         MultiplyByMatrix(&hms.field2screen, &fieldpos[2 + i], &screenpos[3 - i]);
      }
      // FillPolygon(screenpos, 4);
      TextureRectangle(screenpos, mapptr, 1.0);

      /* Parapet */
      fieldpos[0].z = fieldpos[1].z = wall->pos[0].z + 39;
      fieldpos[2].z = fieldpos[3].z = wall->pos[0].z + 50;
      for (k = 0; k < 3; k++) {
         if (wall->align == XALIGNED) {
	    fieldpos[0].x = fieldpos[3].x = wall->pos[0].x - k * 20.0;
	    fieldpos[1].x = fieldpos[2].x = fieldpos[0].x - 10.0;
	 }
         else {
	    fieldpos[0].y = fieldpos[3].y = wall->pos[0].y - k * 20.0;
	    fieldpos[1].y = fieldpos[2].y = fieldpos[0].y - 10.0;
         }
         for (i = 0; i < 4; i++)
            MultiplyByMatrix(&hms.field2screen, &fieldpos[i], &screenpos[i]);
         // FillPolygon(screenpos, 4);
         TextureRectangle(screenpos, mapptr, 0.2);
      }
   }
}

/*--------------------------------------------------------------*/
/* Redraw the graphics for the active state			*/
/*--------------------------------------------------------------*/

void update_active_state()
{
   fpoint fieldpos, bpos, screenpos[12];
   fpoint mark_vec[3] = {
	{2.0,  0.0, -1.0},
	{10.0,  3.0, -1.0},
	{10.0, -3.0, -1.0}
   };
   fpoint circ_vec[12] = {
	{ 0.0,  8.0, 0.0},
	{-4.0,  6.9, 0.0},
	{-6.9,  4.0, 0.0},
	{-8.0,  0.0, 0.0},
	{-6.9, -4.0, 0.0},
	{-4.0, -6.9, 0.0},
	{ 0.0, -8.0, 0.0},
	{ 4.0, -6.9, 0.0},
	{ 6.9, -4.0, 0.0},
	{ 8.0,  0.0, 0.0},
	{ 6.9,  4.0, 0.0},
	{ 4.0,  6.9, 0.0}
   };
   fpoint ground_vec[4] = {
        {-1000.0, -1300.0, -3.0},
        { 1000.0, -1300.0, -3.0},
        { 1000.0,  1000.0, -3.0},
        {-1000.0,  1000.0, -3.0}
   };

   int i, j, k, eleft, total;
   float size;
   fpoint forig;
   char pstr[50];
   Enemy *enemy;
   Boulder *boulder;
   Bonus *bonus;
   Target *target;
   Castle *wall;

   // Main application drawing routine goes here.

   render_init();

   // SetForeground(LIGHT_GREEN);

   for (i = 0; i < 4; i++)
      MultiplyByMatrix(&hms.field2screen, &ground_vec[i], &screenpos[i]);
   TextureRectangle(screenpos, &hms.grass_map, 16.0);
   // FillPolygon(screenpos, 4);

   // Draw dead enemy markers

   for (i = 0; i < hms.num_enemies; i++) {
      enemy = hms.enemies + i;
      if (enemy->state == DEAD) {
	 SetForeground(DARK_RED);
	 for (k = 0; k < 4; k++) {
 	    for (j = 0; j < 3; j++) {
	       if (k == 1 || k == 3) {
	          fieldpos.x = mark_vec[j].y;
	          fieldpos.y = mark_vec[j].x;
	       }
	       else {
	          fieldpos.x = mark_vec[j].x;
	          fieldpos.y = mark_vec[j].y;
	       }
	       if (k == 2) fieldpos.x = -fieldpos.x;
	       if (k == 3) fieldpos.y = -fieldpos.y;

	       fieldpos.z = mark_vec[j].z;
	       fieldpos.x += enemy->pos.x;
	       fieldpos.y += enemy->pos.y;
	       fieldpos.z += enemy->pos.z;
               MultiplyByMatrix(&hms.field2screen, &fieldpos, &screenpos[j]);
	    }
            FillPolygon(screenpos, 3);
	 }
      }
   }

   // Draw targets

   for (j = 0; j < hms.num_targets; j++) {
      target = hms.targets + j;

      switch(target->type) {
	 case SHIELD:
	    SetForeground(LIGHT_BLUE);
	    break;
	 case BOMB:
	    SetForeground(ORANGE);
	    break;
	 case LIFE:
	    SetForeground(YELLOW);
	    break;
	 case BONUS:
	    SetForeground(PURPLE);
	    break;
      }

      switch(target->state) {
	 case GONE:
	    size = 0.0;
	    break;
	 default:
	    bpos = target->pos;
	    size = 2.0;
	    break;
      }

      if (size > 0) {
         for (i = 0; i < 12; i++) {
            fieldpos.x = bpos.x + size * circ_vec[i].x;
            fieldpos.y = bpos.y + size * circ_vec[i].y;
            fieldpos.z = -2;
            MultiplyByMatrix(&hms.field2screen, &fieldpos, &screenpos[i]);
         }
         FillPolygon(screenpos, 12);
      }
   }

   for (j = 0; j < hms.num_pieces; j++) 
      draw_castle(hms.pieces + j);

   for (j = 0; j < hms.num_enemies; j++) 
      draw_enemy(hms.enemies + j, circ_vec);

   draw_catapult(hms.catapult, circ_vec);

   for (j = 0; j < hms.num_boulders; j++) 
      draw_boulder(hms.boulders + j, circ_vec);

   // Draw bonus scores

   for (i = 0; i < hms.num_bonuses; i++) {
      bonus = hms.bonuses + i;
      if (!compare_time(&bonus->timeout)) {
	 SetForeground(DARK_BLUE);
         sprintf(pstr, "+%d", bonus->value);
         DrawString(bonus->pos, pstr, strlen(pstr), 0);
      }
   }

   // Draw bars at the top representing the lives we have left.

   SetForeground(YELLOW);
   forig.y = 12;
   forig.z = 100;
   for (i = 0; i < hms.lives - 1; i++) {
       forig.x = 10 + i * 7;
       FillRectangle(forig, 5, 15); 
   }

   // Draw bars at the top representing the bombs we have

   SetForeground(ORANGE);
   for (i = 0; i < hms.bombs; i++) {
       forig.x = 550 + i * 7;
       FillRectangle(forig, 5, 15); 
   }

   // Continue bars at the top representing the wall pieces we have

   SetForeground(LIGHT_BLUE);
   for (i = 0; i < hms.shields; i++) {
       forig.x = 550 + (hms.bombs + i) * 7;
       FillRectangle(forig, 5, 15); 
   }

   // Add number of enemies left, and the score

   SetForeground(PURPLE);
   sprintf(pstr, "Score: %d", hms.score);
   forig.x = 130;
   forig.z = 100;
   forig.y = 20;
   DrawString(forig, pstr, strlen(pstr), 0);
   for (eleft = 0, i = 0; i < hms.num_enemies; i++)
      if (hms.enemies[i].state >= 0)
	 eleft++;
   sprintf(pstr, "Enemies left: %d", eleft);
   forig.x = 250;
   DrawString(forig, pstr, strlen(pstr), 0);

   sprintf(pstr, "Level: %d", hms.level);
   forig.x = 430;
   DrawString(forig, pstr, strlen(pstr), 0);
}

/*--------------------------------------------------------------*/
/* Redraw the buffer graphics					*/
/*--------------------------------------------------------------*/

void update_window()
{
   switch(hms.gamestate) {
       case SPLASH:
	  draw_title();
	  draw_splash();
	  break;
       case SCORES:
	  draw_title();
	  draw_scores();
	  break;
       case INSTRUCTIONS:
	  draw_title();
	  draw_help();
	  break;
       case RETRY:
	  update_active_state();
	  draw_retry();
	  break;
       case INTERLEVEL:
	  update_active_state();
	  draw_interlevel();
	  break;
       case GAMEOVER:
	  update_active_state();
	  draw_gameover();
	  break;
       case WINNER:
	  update_active_state();
	  draw_winner();
	  break;
       case ACTIVE:
	  update_active_state();
	  break;
   }
   if (hms.paused > 0) draw_paused();
   copy_window();
}

/*--------------------------------------------------------------*/
/* Initialize the game state---first time through (especially	*/
/* for malloc'd memory)						*/
/*--------------------------------------------------------------*/

void initialize_game()
{
    read_high_scores();

    /* Seed random number generator and set current time at start */
    gettimeofday(&hms.now, NULL);
    srand((unsigned int)hms.now.tv_usec);
    hms.boulders = NULL;
    hms.enemies = NULL;
    hms.bonuses = NULL;
    hms.num_boulders = 0;
    hms.num_enemies = 0;
    hms.num_bonuses = 0;
    hms.num_targets = 0;
    hms.num_pieces = 0;
    hms.paused = 0;
    hms.catapult = (Catapult *)Tcl_Alloc(sizeof(Catapult));
    hms.cur_color = (XColor *)NULL;
}

/*--------------------------------------------------------------*/
/* Restart the game						*/
/*--------------------------------------------------------------*/

void restart_game(int transx, int transy, int anglex, int angley, int anglez)
{
    fpoint tmpvector;
    Matrix tmp, tmp2;
    int i;

    hms.catapult->state = -1;		/* -1 = NOT READY */
    hms.catapult->angle = 0;
    hms.catapult->pull = 0;
    hms.active_pull = 0;
    hms.active_rotate = 0;
    hms.catapult->pos.x = 0;
    hms.catapult->pos.y = 0;
    hms.catapult->pos.z = BASELINE;
    catapult_corners();

    do_pull(0);				/* Compute arm/boulder position */

    /* Generate transformation matrix */

    IdentityMatrix(&tmp);

    RotateMatrixX(&tmp, &tmp2, anglex);
    RotateMatrixY(&tmp2, &tmp, angley);
    RotateMatrixZ(&tmp, &tmp2, anglez);

    tmpvector.x = (float)transx;
    tmpvector.y = (float)transy;
    tmpvector.z = 0.0;
    TranslateMatrix(&tmp2, &hms.field2screen, &tmpvector);

    InvertMatrix(&hms.field2screen, &hms.screen2field);

    hms.gamestate = SPLASH;
    hms.level = 1;
    hms.score = 0;
    hms.lives = 2;
    hms.bombs = 1;
    hms.shields = 0;

    if (hms.num_boulders > 0) {
       Tcl_Free(hms.boulders);
       hms.num_boulders = 0;
       hms.boulders = NULL;
    }
    if (hms.num_enemies > 0) {
       Tcl_Free(hms.enemies);
       hms.num_enemies = 0;
       hms.enemies = NULL;
    }
    if (hms.num_pieces > 0) {
       Tcl_Free(hms.pieces);
       hms.num_pieces = 0;
       hms.pieces = NULL;
    }

    gettimeofday(&hms.now, NULL);
    set_future_time(&hms.timeout, 3.0);
}

/*------------------------------------------------------*/
