/*----------------------------------------------------------------------*/
/* helpme.h 								*/
/* Copyright (c) 2009  Tim Edwards					*/
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*      written by Tim Edwards, 6/21/09    				*/
/*----------------------------------------------------------------------*/

#ifndef _HELPME_H
#define _HELPME_H

#include <time.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <tk.h>

#include <GL/gl.h>
#include <GL/glx.h>

#include "colordefs.h"

#ifndef HAVE_U_CHAR
typedef unsigned char	u_char;
typedef unsigned short	u_short;
typedef unsigned int	u_int;
typedef unsigned long	u_long;
#endif

/*----------------------------------------------------------------------*/
/* Deal with 32/64 bit processors based on autoconf results.		*/
/*----------------------------------------------------------------------*/

#if SIZEOF_VOID_P == SIZEOF_UNSIGNED_INT
#define Number(a)		(void *)((u_int) a)
typedef u_int pointertype;
#elif SIZEOF_VOID_P == SIZEOF_UNSIGNED_LONG
#define Number(a)		(void *)((u_long) a)
typedef u_long pointertype;
#else
ERROR: Cannot compile without knowing the size of a pointer.  See helpme.h.
#endif

/*----------------------------------------------------------------------*/
/* Definition shortcuts 						*/
/*----------------------------------------------------------------------*/

#define abs(a)			((a) < 0 ? -(a) : (a))
#define sign(a)			((a) <= 0 ? -1 : 1)
#ifndef min
#define max(a,b)		((a) < (b) ? (b) : (a))
#define min(a,b)		((a) < (b) ? (a) : (b))
#endif

/*----------------------------------------------------------------------*/
/* Geometry								*/
/*----------------------------------------------------------------------*/

#define RAD_TO_DEG 570.29578		/* Units are 1/10 degree */
#define DEG_TO_RAD 0.00174533		/* Units are 1/10 degree */

typedef struct {
   float x;
   float y;
   float z;
} fpoint;		/* Point in 3 dimensions */

typedef struct {
   float a, b, c;
   float d, e, f;
   float g, h, i;
   float j, k, l;
} Matrix;		/* Transformation matrix */
			/* a  b  c  0 */
			/* d  e  f  0 */
			/* g  h  i  0 */
			/* j  k  l  1 */

/*----------------------------------------------------------------------*/
/* Geometric constants
/*----------------------------------------------------------------------*/

#define BASELINE  8	/* Height of catapult base above ground */

/*----------------------------------------------------------------------*/
/* Graphics								*/
/*----------------------------------------------------------------------*/

typedef struct {
   int width;
   int height;
   GLuint id;
   GLubyte *data;
} TextureMap;

/*----------------------------------------------------------------------*/
/* Game states								*/
/*----------------------------------------------------------------------*/

enum {SPLASH, SCORES, INSTRUCTIONS, ACTIVE, RETRY, INTERLEVEL, GAMEOVER,
	WINNER};

/*----------------------------------------------------------------------*/
/* Target types								*/
/*----------------------------------------------------------------------*/

enum {BONUS, SHIELD, BOMB, LIFE, NUM_TARGET_TYPES};

/*----------------------------------------------------------------------*/
/* Target states							*/
/*----------------------------------------------------------------------*/

enum {GONE, UNCLAIMED, TARGETED};

/*----------------------------------------------------------------------*/
/* Enemy states	and strategies						*/
/*----------------------------------------------------------------------*/

enum {DEAD = -1, STOPPED, MOVING};
enum {ADVANCE, ARC_LEFT, ARC_RIGHT, SCATTER, RETREAT, BACK_OFF, GET_TARGET};

/*----------------------------------------------------------------------*/
/* Castle wall types
/*----------------------------------------------------------------------*/

enum {XALIGNED, YALIGNED};
enum {UNPLACED, BUILT, DESTROYED};

/*----------------------------------------------------------------------*/
/* Structure for tracking targets					*/
/*----------------------------------------------------------------------*/

typedef struct {
   fpoint	pos;
   int		type;
   int		state;
} Target;

/*----------------------------------------------------------------------*/
/* Structure for tracking bonus scores					*/
/*----------------------------------------------------------------------*/

typedef struct {
   fpoint		pos;
   int			value;
   struct timeval	timeout;	// Negative timeout means inactive
} Bonus;

/*----------------------------------------------------------------------*/
/* Structure for each enemy position and state				*/
/*----------------------------------------------------------------------*/

typedef struct {
   fpoint 		pos;
   int			angle;		/* 0 to 360 in 10ths */
   char			state;
   u_char		strategy;
   u_char		flags;		// SHIELD and/or BOMB carrier
   struct timeval	timeout;	// Time to reevaluate strategy
} Enemy;

/*----------------------------------------------------------------------*/
/* Structure for each castle wall piece					*/
/*----------------------------------------------------------------------*/

typedef struct {
   fpoint	pos[2];
   u_char	align;
   u_char	type;
   u_char	strength;	// Wall is destroyed when strength -> 0
} Castle;

/*----------------------------------------------------------------------*/
/* Structure for the catapult						*/
/*----------------------------------------------------------------------*/

typedef struct {
   int		state;		/* Ready, Firing, Reloading	*/
   int		angle;		/* 0 to 360 in 10ths		*/
   int		pull;		/* 0 to 45			*/
   fpoint	pos;		/* Catapult position		*/
   fpoint	corner[4];	/* Corner positions		*/
} Catapult;

/*----------------------------------------------------------------------*/
/* Structure for tracking boulders after they have landed		*/
/* If "hits" is -1, then the boulder is a "bomb" type.			*/
/*----------------------------------------------------------------------*/

typedef struct {
   int			state;		/* Bounce/move state		*/
   float		zvelocity;	/* Boulder vertical velocity	*/
   float		fvelocity;	/* Boulder forward velocity	*/
   float		delv;		/* Delta vertical velocity	*/
   fpoint		pos;		/* Position of boulder in 3D	*/
   int			angle;		/* Copied from catapult->angle	*/
   int			hits;		/* Hits per boulder (for bonus)	*/
   struct timeval	timeout;	/* Time to remove		*/
} Boulder;

/*----------------------------------------------------------------------*/
/*----------------------------------------------------------------------*/

typedef struct {
   char    name[20];
   char    date[20];
   int     score;
   u_short level;
   u_char  win;
} HighScore;

/*----------------------------------------------------------------------*/
/* A convenient structure for holding globally-accessible data		*/
/*----------------------------------------------------------------------*/

typedef struct {

   // Window/graphics data

   Tk_Window	topwin;			/* Top-level window of the app  */
   Tk_Window	area;			/* Window we draw into		*/
   Window	window;			/* X11 handle of the window	*/
   short	width, height;		/* Width/height of window */
   XColor	colors[NUMBER_OF_COLORS];
   XColor	*cur_color;

   Matrix	screen2field;		/* Forward transform		*/
   Matrix	field2screen;		/* Reverse transform		*/

   XFontStruct	*fonts[2];		/* For text width computation	*/

   TextureMap	grass_map;		/* Grass texture map		*/
   TextureMap	wood_map;		/* Wood texture map		*/
   TextureMap	stone_map;		/* Stone texture map		*/
   TextureMap	stone2_map;		/* Stone texture map (shaded)	*/
   TextureMap	blue_map;		/* Splash screen texture map 	*/

   // Game data

   HighScore		high_scores[10];
   char			name[20];	/* Player's name for score list	*/
   u_char		active_pull;	/* 0 or 1 */			
   char			active_roll;	/* -1, 0, or 1 */			
   char			active_rotate;	/* -1, 0, or 1 */			
   int			gamestate;	/* Splash screen, etc.		*/
   struct timeval	now;		/* Current time			*/
   struct timeval	timeout;	/* Timeout timer for interlevel	*/
   time_t		paused;		/* Time game was paused		*/
   int			lives;
   int			score;
   int			level;
   int			bombs;
   int			shields;	/* Number of castle pieces	*/
   int			num_enemies;	/* Number of enemies		*/
   Enemy		*enemies;
   int			num_boulders;	/* Number of boulders		*/
   Boulder		*boulders;	/* Boulders already landed	*/
   int			num_bonuses;	/* Number of bonus markers	*/
   Bonus		*bonuses;	/* bonus markers		*/
   int			num_targets;	/* Number of target markers	*/
   Target		*targets;
   int			num_pieces;	/* Number of castle pieces	*/
   Castle		*pieces;
   Catapult		*catapult;
} Globaldata;

/*----------------------------------------------------------------------*/
/* External variable declarations					*/
/*----------------------------------------------------------------------*/

extern Globaldata hms;
extern Tcl_Interp *helpmeinterp;
extern Display *dpy;
extern Tcl_HashTable HelpMeTagTable;
extern Display *dpy;
extern Colormap cmap;
extern GLXContext grXcontext;

/*----------------------------------------------------------------------*/
/* Function prototypes							*/
/*----------------------------------------------------------------------*/

extern int Tk_SimpleObjCmd();
extern int HelpmeTagCallback();
extern int HelpmeInternalTagCall(Tcl_Interp *, int, ...);

extern int helpme_config();
extern int helpme_redraw();
extern int helpme_recalc();
extern int helpme_nextstate();
extern int helpme_pause();
extern int helpme_launch();
extern int helpme_quit();
extern int helpme_view();
extern int helpme_pull();
extern int helpme_bomb();
extern int helpme_rotate();
extern int helpme_roll();
extern int helpme_position();
extern int helpme_tag();
extern int helpme_build();
extern int helpme_username();
extern int helpme_nextlevel();

extern void do_rotate(int);
extern void do_roll(int);
extern void do_pull(int);
extern void do_bomb();
extern void do_launch(int, int);
extern void get_field_position(int, int, fpoint *);
extern void restart_game(int, int, int, int, int);
extern void restart_game();
extern void draw_splash();
extern void draw_scores();
extern void draw_help();
extern void draw_title();
extern u_char set_wall_pos();
extern void render_init();
extern void copy_window();
extern void helpme_resizearea(ClientData, XEvent *);
extern void helpme_drawarea(ClientData, XEvent *);
extern int GUI_init(Tcl_Interp *, int, Tcl_Obj *CONST objv[]);
extern int new_window(Tcl_Interp *, Tk_Window);
extern int helpme_start(ClientData, Tcl_Interp *, int, Tcl_Obj *CONST objv[]);

extern u_char new_texture(char *, TextureMap *);
extern void shade_texture(TextureMap *, TextureMap *, float);

extern char *Tcl_Strdup(const char *);

extern void read_high_scores();
extern void save_high_scores();
extern u_char check_high_score();
extern void record_high_score(u_char);
extern u_char check_victory();

extern void FillRectangle(fpoint, int, int);
extern void FillArc(fpoint, int, fpoint *);
extern void FillPolygon(fpoint *, int);
extern void DrawString(fpoint, char *, int, int);
extern void DrawLine(fpoint *, fpoint *);
extern void DrawLines(fpoint *, int, u_char);
extern void TextureRectangle(fpoint *, TextureMap *, float);


extern void IdentityMatrix(Matrix *mat);
extern void InvertMatrix(Matrix *, Matrix *);
extern void MultiplyByMatrix(Matrix *, fpoint *, fpoint *);
extern void MultiplyMatrices(Matrix *, Matrix *, Matrix *);
extern void RotateMatrixY(Matrix *, Matrix *, float);
extern void RotateMatrixX(Matrix *, Matrix *, float);
extern void RotateMatrixZ(Matrix *, Matrix *, float);
extern void TranslateMatrix(Matrix *, Matrix *, fpoint *);
extern float finddist(fpoint *, fpoint *, fpoint *);
extern float fsqlinelen(fpoint *, fpoint *);
extern u_char test_intersect(fpoint *, fpoint *);

/*-----------------------*/
/* Tcl 8.4 compatibility */
/*-----------------------*/

#ifndef CONST84
#define CONST84
#endif

/*----------------------------------------------------------------------*/
/* Deal with systems which don't define va_copy().			*/
/*----------------------------------------------------------------------*/

#ifndef HAVE_VA_COPY
  #ifdef HAVE___VA_COPY
    #define va_copy(a, b) __va_copy(a, b)
  #else
    #define va_copy(a, b) a = b
  #endif
#endif

/*----------------------------------------------------------------------*/

#endif /* _HELPME_H */
