/*--------------------------------------------------------------*/
/* helpmewrap.c:						*/
/*	Tcl module generator for helpme				*/
/*--------------------------------------------------------------*/

#ifdef TCL_WRAPPER

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "helpme.h"

Tcl_Interp *helpmeinterp;

/*------------------------------------------------------*/
/* Create command callback functions in Tcl/Tk    	*/    
/*------------------------------------------------------*/    

typedef struct {
   const char	*cmdstr;
   void		(*func)();
} cmdstruct;

static cmdstruct helpme_commands[] =
{
   {"config", (void *)helpme_config},
   {"redraw", (void *)helpme_redraw},
   {"recalc", (void *)helpme_recalc},
   {"launch", (void *)helpme_launch},
   {"start", (void *)helpme_start},
   {"pause", (void *)helpme_pause},
   {"nextstate", (void *)helpme_nextstate},
   {"nextlevel", (void *)helpme_nextlevel},
   {"tag", (void *)helpme_tag},
   {"view", (void *)helpme_view},
   {"position", (void *)helpme_position},
   {"pull", (void *)helpme_pull},
   {"build", (void *)helpme_build},
   {"bomb", (void *)helpme_bomb},
   {"rotate", (void *)helpme_rotate},
   {"roll", (void *)helpme_roll},
   {"username", (void *)helpme_username},
   {"quit", (void *)helpme_quit},
   {"", NULL}  /* sentinel */
};

/*--------------------------------------------------------------*/
/* helpme commands						*/
/*--------------------------------------------------------------*/

/*--------------------------------------------------------------*/
/* function: helpme_config					*/
/* command:  config 						*/
/*--------------------------------------------------------------*/

int helpme_config(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
   static char *subCmds[] = {"window", "id", NULL};
   enum SubIdx {WindowIdx, IDIdx};
   int idx, result;
   Tk_Window tkwind;

   if (objc == 1) {
      Tcl_WrongNumArgs(interp, 1, objv, "option ?arg ...?");
      return TCL_ERROR;
   }
   else if ((result = Tcl_GetIndexFromObj(interp, objv[1],
	(CONST84 char **)subCmds, "option", 0, &idx)) != TCL_OK) {
      return result;
   }
   switch(idx) {
       case WindowIdx:
	  if (objc == 2) {
	     if (hms.area == (Tk_Window)NULL)
		Tcl_SetResult(interp, "There is no target window.", NULL);
	     else
                Tcl_AppendElement(interp, Tk_PathName(hms.area));
	  }
	  else if (objc == 3) {
	     tkwind = Tk_NameToWindow(interp, Tcl_GetString(objv[2]), hms.topwin);
	     if (tkwind == (Tk_Window)NULL) {
		Tcl_SetResult(interp, "No such window!", NULL);
	 	return TCL_ERROR;
	     }
	     result = new_window(interp, tkwind);
	     if (result == -2) {
		Tcl_SetResult(interp, "Bad window for drawing!", NULL);
	 	return TCL_ERROR;
	     }
	  }
	  else {
	     Tcl_WrongNumArgs(interp, 2, objv, "(no arguments)");
	     return TCL_ERROR;
	  }
	  break;

       case IDIdx:
	  if (objc == 2)
             Tcl_SetObjResult(interp, Tcl_NewIntObj(hms.window));
	  else {
	     Tcl_WrongNumArgs(interp, 2, objv, "(no arguments)");
	     return TCL_ERROR;
	  }
	  break;
   }
   return HelpMeTagCallback(interp, objc, objv);
}

/*--------------------------------------------------------------*/
/* function: helpme_position					*/
/* command:  position						*/
/* Return the field coordinate at z = 0 for the given screen	*/
/* coordinate in (x, y).					*/
/*--------------------------------------------------------------*/

int helpme_position(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
    int x, y, result;
    fpoint fp;
    Tcl_Obj *lobj;

    if (objc != 3) {
       Tcl_WrongNumArgs(interp, 1, objv, "<x> <y>");
       return TCL_ERROR;
    }

    result = Tcl_GetIntFromObj(interp, objv[1], &x);
    if (result != TCL_OK) return result;
    result = Tcl_GetIntFromObj(interp, objv[2], &y);
    if (result != TCL_OK) return result;

    get_field_position(x, y, &fp);

    lobj = Tcl_NewListObj(0, NULL);

    Tcl_ListObjAppendElement(interp, lobj, Tcl_NewDoubleObj((double)fp.x));
    Tcl_ListObjAppendElement(interp, lobj, Tcl_NewDoubleObj((double)fp.y));

    Tcl_SetObjResult(interp, lobj);
    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* function: helpme_pause					*/
/* command:  pause						*/
/* Pause the game (pause mode)					*/
/*--------------------------------------------------------------*/

int helpme_pause(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
    Bonus *bonus;
    Boulder *boulder;
    Enemy *enemy;
    int result, i;
    time_t adder;

    if (objc != 1) {
       Tcl_WrongNumArgs(interp, 1, objv, "(no arguments)");
       return TCL_ERROR;
    }
    if (hms.paused == 0)
        hms.paused = hms.now.tv_sec;
    else {
	// Adjust timeouts
	adder = hms.now.tv_sec - hms.paused;
        hms.paused = 0;
	hms.timeout.tv_sec += adder;
	for (i = 0; i < hms.num_boulders; i++) {
	   boulder = hms.boulders + i;
	   boulder->timeout.tv_sec += adder;
	}
	for (i = 0; i < hms.num_bonuses; i++) {
	   bonus = hms.bonuses + i;
	   bonus->timeout.tv_sec += adder;
	}
	for (i = 0; i < hms.num_enemies; i++) {
	   enemy = hms.enemies + i;
	   enemy->timeout.tv_sec += adder;
	}
    }
    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* function: helpme_build					*/
/* command:  build						*/
/* Turn a reserved shield into a castle piece			*/
/*--------------------------------------------------------------*/

int helpme_build(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
    int result, i;
    float x, y;
    Castle *wall;

    if (objc != 1) {
       Tcl_WrongNumArgs(interp, 1, objv, "(no arguments)");
       return TCL_ERROR;
    }
    
    /* If we are tracking an unbuilt piece, then build it.	*/
    /* Otherwise, start tracking a new unbuilt piece.		*/

    for (i = 0; i < hms.num_pieces; i++) {
       wall = hms.pieces + i;
       if (wall->type == UNPLACED) {
	  if (set_wall_pos() == 0) {
	     wall->type = BUILT;
	     if (check_victory() == 1) {
	        // End the game---you won!
		hms.gamestate = WINNER;
	     }
	  }
	  return TCL_OK;
       }
    }
	
    if (hms.shields > 0) {
       hms.shields--;

       /* Recycle memory from destroyed wall structures */
       for (i = 0; i < hms.num_pieces; i++) {
	  wall = hms.pieces + i;
          if (wall->type == DESTROYED)
	     break;
       }

       if (i == hms.num_pieces) {
          hms.num_pieces++;
          if (hms.num_pieces == 1)
	     hms.pieces = (Castle *)Tcl_Alloc(sizeof(Castle));
          else
	     hms.pieces = (Castle *)Tcl_Realloc(hms.pieces,
			hms.num_pieces * sizeof(Castle));
          wall = hms.pieces + hms.num_pieces - 1;
       }
       wall->type = UNPLACED;
       wall->strength = (u_char)255;
	
       set_wall_pos();
    }
    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* function: helpme_bomb					*/
/* command:  bomb						*/
/* Turn a loaded boulder into a bomb				*/
/*--------------------------------------------------------------*/

int helpme_bomb(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
    int result;

    if (objc != 1) {
       Tcl_WrongNumArgs(interp, 1, objv, "(no arguments)");
       return TCL_ERROR;
    }
    
    make_bomb();
    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* function: helpme_launch					*/
/* command:  launch						*/
/* Load or Launch a boulder					*/
/*--------------------------------------------------------------*/

int helpme_launch(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
    int pull, angle, result;

    if (objc == 1) {
	do_launch(-1, -1);
	return TCL_OK;
    }
    else if (objc != 3) {
       Tcl_WrongNumArgs(interp, 1, objv, "[<pull> <angle>]");
       return TCL_ERROR;
    }
    
    result = Tcl_GetIntFromObj(interp, objv[1], &pull);
    if (result != TCL_OK) return result;
    result = Tcl_GetIntFromObj(interp, objv[2], &angle);
    if (result != TCL_OK) return result;

    do_launch(pull, angle);
    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* function: helpme_recalc					*/
/* command:  recalc						*/
/* Calculate new positions for everything			*/
/*--------------------------------------------------------------*/

int helpme_recalc(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
    recalculate();
    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* function: helpme_redraw					*/
/* command:  redraw						*/
/*--------------------------------------------------------------*/

int helpme_redraw(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
   if (objc != 1) {
      Tcl_WrongNumArgs(interp, 1, objv, "(no options)");
      return TCL_ERROR;
   }
   update_window();
   return TCL_OK;
}

/*--------------------------------------------------------------*/
/* function: helpme_quit					*/
/* command:  quit						*/
/*--------------------------------------------------------------*/

int helpme_quit(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
   if (objc != 1) {
      Tcl_WrongNumArgs(interp, 1, objv, "(no options)");
      return TCL_ERROR;
   }

   if (hms.gamestate == ACTIVE)
      hms.gamestate = GAMEOVER;
   else
      Tcl_Exit(0);

   return TCL_OK;
}

/*--------------------------------------------------------------*/
/* function: helpme_view					*/
/* command:  view						*/
/*--------------------------------------------------------------*/

int helpme_view(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
    int x, y, theta, phi, psi, result;

    if (objc != 6) {
	Tcl_WrongNumArgs(interp, 1, objv, "<x> <y> <theta> <phi> <psi>");
	return TCL_ERROR;
    }

    result = Tcl_GetIntFromObj(interp, objv[1], &x);
    if (result != TCL_OK) return result;
    result = Tcl_GetIntFromObj(interp, objv[2], &y);
    if (result != TCL_OK) return result;
    result = Tcl_GetIntFromObj(interp, objv[3], &theta);
    if (result != TCL_OK) return result;
    result = Tcl_GetIntFromObj(interp, objv[4], &phi);
    if (result != TCL_OK) return result;
    result = Tcl_GetIntFromObj(interp, objv[5], &psi);
    if (result != TCL_OK) return result;

    restart_game(x, y, theta, phi, psi);

    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* function: helpme_nextlevel					*/
/* command:  nextlevel						*/
/*--------------------------------------------------------------*/

int helpme_nextlevel(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 1, objv, "(no arguments)");
	return TCL_ERROR;
    }
    hms.gamestate = INTERLEVEL;
    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* function: helpme_nextstate					*/
/* command:  nextstate						*/
/*--------------------------------------------------------------*/

int helpme_nextstate(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
    if (objc != 1) {
	Tcl_WrongNumArgs(interp, 1, objv, "(no arguments)");
	return TCL_ERROR;
    }

    switch(hms.gamestate) {
       case SPLASH:
       case INSTRUCTIONS:
       case SCORES:
	  restart_game(200, 200, 320, 10, 320);
	  generate_new_enemies();
	  hms.gamestate = ACTIVE;
	  break;
       case GAMEOVER:
       case WINNER:
	  hms.gamestate = SPLASH;
	  set_future_time(&hms.timeout, 3.0);
	  break;
    }
    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* function: helpme_pull					*/
/* command:  pull						*/
/*--------------------------------------------------------------*/

int helpme_pull(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
    int amount, result;

    if (objc == 1) {
	Tcl_SetObjResult(interp, Tcl_NewIntObj(hms.catapult->pull));
	return TCL_OK;
    }
    else if (objc != 2) {
	Tcl_WrongNumArgs(interp, 1, objv, "[<amount>]");
	return TCL_ERROR;
    }

    /* Check for options "enable" and "disable" */
    if (!strncmp(Tcl_GetString(objv[1]), "ena", 3)) {
	hms.active_pull = 1;
    }
    else if (!strncmp(Tcl_GetString(objv[1]), "dis", 3)) {
	hms.active_pull = 0;
	return TCL_OK;
    }
    else {
        result = Tcl_GetIntFromObj(interp, objv[1], &amount);
	if (result != TCL_OK) return result;
    }

    if (hms.active_pull != 0) do_pull(amount);

    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* function: helpme_username					*/
/* command:  username <name>					*/
/*--------------------------------------------------------------*/

int helpme_username(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
    u_char win;

    if (objc != 2) {
	Tcl_WrongNumArgs(interp, 1, objv, "<name>");
	return TCL_ERROR;
    }
    strncpy(hms.name, Tcl_GetString(objv[1]), 19);

    // If we're in the WINNER state, then sort and save the high score

    if (hms.gamestate == WINNER) {
       if (check_high_score() == 1) {
          win = check_victory();
          record_high_score(win);
          save_high_scores();
       }
       hms.gamestate = GAMEOVER;
    }

    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* function: helpme_roll					*/
/* command:  roll [amount]	or				*/
/*	     roll forward|backward|disable			*/
/*--------------------------------------------------------------*/

int helpme_roll(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
    int amount, result, i;
    float ccos, csin, famt;
    fpoint savepos;
    char *obj1str;

    if ((objc != 1) && (objc != 2)) {
	Tcl_WrongNumArgs(interp, 1, objv, "[<amount>]");
	return TCL_ERROR;
    }
    else if (objc == 1) {
	do_roll(20);
	return TCL_OK;
    }
    else
        obj1str = Tcl_GetString(objv[1]);

    if (!strncmp(obj1str, "forw", 4)) {
	hms.active_roll = 1;
    }
    else if (!strncmp(obj1str, "back", 4)) {
	hms.active_roll = -1;
    }
    else if (!strncmp(obj1str, "dis", 3)) {
	hms.active_roll = 0;
    }
    else {
        result = Tcl_GetIntFromObj(interp, objv[1], &amount);
	if (result != TCL_OK) return result;
        if (hms.active_roll == 0) do_roll(amount);
    }
    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* function: helpme_rotate					*/
/* command:  rotate						*/
/*--------------------------------------------------------------*/

int helpme_rotate(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
    int amount, result, nullint, x, y, relative = 0, diff;
    Window nullwin;
    u_int  nullui;
    fpoint fp;
    float angle;
    u_char doena = 0;

    if (objc == 1) {
	Tcl_SetObjResult(interp, Tcl_NewIntObj(hms.catapult->angle));
	return TCL_OK;
    }
    else if ((objc == 3) && !strncmp(Tcl_GetString(objv[2]), "to", 2)) {
	XQueryPointer(dpy, Tk_WindowId(hms.area), &nullwin, &nullwin,
		&nullint, &nullint, &x, &y, &nullui);
        get_field_position(x, y, &fp);
	angle = RAD_TO_DEG * (float)atan2((double)(fp.y - hms.catapult->pos.y),
			(double)(fp.x - hms.catapult->pos.x));
        relative = 1;
    }
    else if (objc != 2) {
	Tcl_WrongNumArgs(interp, 1, objv, "[<amount>]");
	return TCL_ERROR;
    }

    /* Check for options "enable" and "disable" */
    if (!strncmp(Tcl_GetString(objv[1]), "ena", 3)) {
	amount = 1;
	doena = 1;
    }
    else if (!strncmp(Tcl_GetString(objv[1]), "dis", 3)) {
	hms.active_rotate = 0;
	return TCL_OK;
    }
    else {
        result = Tcl_GetIntFromObj(interp, objv[1], &amount);
	if (result != TCL_OK) return result;
    }

    if (relative) {
       amount = abs(amount);
       diff = (int)angle - hms.catapult->angle;
       normalize_angle(&diff);
       if (diff < 0) amount = -amount;
    }
    if (doena == 0) do_rotate(amount);
    else hms.active_rotate = (char)amount;

    return TCL_OK;
}

/*--------------------------------------------------------------*/
/* Initialization procedure for Tcl/Tk				*/
/*--------------------------------------------------------------*/

int
Helpme_Init(Tcl_Interp *interp)
{
   char command[256];
   int cmdidx, clen;
   Tk_Window tktop;
   char version_string[20];

   /* Interpreter sanity checks */
   if (interp == NULL) return TCL_ERROR;

   /* Remember the interpreter */
   helpmeinterp = interp;

   if (Tcl_InitStubs(interp, "8.1", 0) == NULL) return TCL_ERROR;
   
   strcpy(command, "helpme::");
   clen = strlen(command);

   /* Create the start command */

   tktop = Tk_MainWindow(interp);

   /* Create all of the commands (except "simple") */

   for (cmdidx = 0; helpme_commands[cmdidx].func != NULL; cmdidx++) {
      sprintf(command + clen, "%s", helpme_commands[cmdidx].cmdstr);
      Tcl_CreateObjCommand(interp, command, 
		(Tcl_ObjCmdProc *)helpme_commands[cmdidx].func,
		(ClientData)tktop, (Tcl_CmdDeleteProc *) NULL);
   }

   /* Command which creates a "simple" window (this is a top-	*/
   /* level command, not one in the helpme namespace which	*/
   /* is why I treat it separately from the other commands).	*/

   Tcl_CreateObjCommand(interp, "simple",
		(Tcl_ObjCmdProc *)Tk_SimpleObjCmd,
		(ClientData)tktop, (Tcl_CmdDeleteProc *) NULL);

   /* Set $HELPME_VERSION and $HELPME_REVISION as Tcl variables */

   sprintf(version_string, "%d", PROG_REVISION);
   Tcl_SetVar(interp, "HELPME_REVISION", version_string, TCL_GLOBAL_ONLY);

   sprintf(version_string, "%g", PROG_VERSION);
   Tcl_SetVar(interp, "HELPME_VERSION", version_string, TCL_GLOBAL_ONLY);

   /* Export the namespace commands */

   Tcl_Eval(interp, "namespace eval helpme namespace export *");
   Tcl_PkgProvide(interp, "HelpMe", version_string);

   Tcl_InitHashTable(&HelpMeTagTable, TCL_STRING_KEYS);
   return TCL_OK;
}

#endif /* TCL_WRAPPER */
