/*--------------------------------------------------------------*/
/* matrix.c:							*/
/*	Routines for "helpme" 3-D graphics arithmetic		*/ 
/*	and other useful geometry routines			*/
/* Copyright (c) 2009  Tim Edwards, Open Circuit Design		*/
/*--------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <math.h>

#include "helpme.h"

/*--------------------------------------------------------------*/
/* Matrix manipulation						*/
/*--------------------------------------------------------------*/

void IdentityMatrix(Matrix *mat)
{
    mat->a = 1.0;
    mat->b = 0.0;
    mat->c = 0.0;
    mat->d = 0.0;
    mat->e = 1.0;
    mat->f = 0.0;
    mat->g = 0.0;
    mat->h = 0.0;
    mat->i = 1.0;
    mat->j = 0.0;
    mat->k = 0.0;
    mat->l = 0.0;
}

/*--------------------------------------------------------------*/
/* This matrix inversion only works for unscaled transforms.	*/
/*--------------------------------------------------------------*/

void InvertMatrix(Matrix *in, Matrix *out)
{
   float det = in->a * in->e * in->i + in->b * in->f * in->g +
	       in->c * in->d * in->h - in->a * in->f * in->h -
	       in->b * in->d * in->i - in->c * in->e * in->g;

   float idet = 1.0 / det;

   out->a = in->e * in->i - in->f * in->h;
   out->b = in->c * in->h - in->b * in->i;
   out->c = in->b * in->f - in->c * in->e;
   out->d = in->f * in->g - in->d * in->i;
   out->e = in->a * in->i - in->c * in->g;
   out->f = in->c * in->d - in->a * in->f;
   out->g = in->d * in->h - in->e * in->g;
   out->h = in->b * in->g - in->a * in->h;
   out->i = in->a * in->e - in->b * in->d;

   out->j = in->d * in->i * in->k + in->e * in->g * in->l + in->f * in->h * in->j
	  - in->d * in->h * in->l - in->e * in->i * in->j - in->f * in->g * in->k;

   out->k = in->a * in->h * in->l + in->b * in->i * in->j + in->c * in->g * in->k
	  - in->a * in->i * in->k - in->b * in->g * in->l - in->c * in->h * in->j;

   out->l = in->a * in->f * in->k + in->b * in->d * in->l + in->c * in->e * in->j
	  - in->a * in->e * in->l - in->b * in->f * in->j - in->c * in->d * in->k;

   out->a *= idet;
   out->b *= idet;
   out->c *= idet;
   out->d *= idet;
   out->e *= idet;
   out->f *= idet;
   out->g *= idet;
   out->h *= idet;
   out->i *= idet;
   out->j *= idet;
   out->k *= idet;
   out->l *= idet;
}

/*--------------------------------------------------------------*/

void MultiplyByMatrix(Matrix *mat, fpoint *in, fpoint *out)
{
   out->x = in->x * mat->a + in->y * mat->d + in->z * mat->g + mat->j; 
   out->y = in->x * mat->b + in->y * mat->e + in->z * mat->h + mat->k; 
   out->z = in->x * mat->c + in->y * mat->f + in->z * mat->i + mat->l; 
}

/*--------------------------------------------------------------*/

void MultiplyMatrices(Matrix *in1, Matrix *in2, Matrix *out)
{
   out->a = in1->a * in2->a + in1->b * in2->d + in1->c * in2->g;
   out->b = in1->a * in2->b + in1->b * in2->e + in1->c * in2->h;
   out->c = in1->a * in2->c + in1->b * in2->f + in1->c * in2->i;
   out->d = in1->d * in2->a + in1->e * in2->d + in1->f * in2->g;
   out->e = in1->d * in2->b + in1->e * in2->e + in1->f * in2->h;
   out->f = in1->d * in2->c + in1->e * in2->f + in1->f * in2->i;
   out->g = in1->g * in2->a + in1->h * in2->d + in1->i * in2->g;
   out->h = in1->g * in2->b + in1->h * in2->e + in1->i * in2->h;
   out->i = in1->g * in2->c + in1->h * in2->f + in1->i * in2->i;
   out->j = in1->j * in2->a + in1->k * in2->d + in1->l * in2->g + in2->j;
   out->k = in1->j * in2->b + in1->k * in2->e + in1->l * in2->h + in2->k;
   out->l = in1->j * in2->c + in1->k * in2->f + in1->l * in2->i + in2->l;
}

/*--------------------------------------------------------------*/

void RotateMatrixY(Matrix *in, Matrix *out, float angle)
{
   Matrix tmp;
   float cosa, sina;

   cosa = (float)cos(angle * 0.017453);
   sina = (float)sin(angle * 0.017453);

   tmp.a = cosa;
   tmp.b = 0;
   tmp.c = -sina;
   tmp.d = 0;
   tmp.e = 1;
   tmp.f = 0;
   tmp.g = sina;
   tmp.h = 0;
   tmp.i = cosa;
   tmp.j = 0;
   tmp.k = 0;
   tmp.l = 0;

   MultiplyMatrices(in, &tmp, out);
}

/*--------------------------------------------------------------*/

void RotateMatrixX(Matrix *in, Matrix *out, float angle)
{
   Matrix tmp;
   float cosa, sina;

   cosa = (float)cos(angle * 0.017453);
   sina = (float)sin(angle * 0.017453);

   tmp.a = cosa;
   tmp.b = -sina;
   tmp.c = 0;
   tmp.d = sina;
   tmp.e = cosa;
   tmp.f = 0;
   tmp.g = 0;
   tmp.h = 0;
   tmp.i = 1;
   tmp.j = 0;
   tmp.k = 0;
   tmp.l = 0;

   MultiplyMatrices(in, &tmp, out);
}

/*--------------------------------------------------------------*/

void RotateMatrixZ(Matrix *in, Matrix *out, float angle)
{
   Matrix tmp;
   float cosa, sina;

   cosa = (float)cos(angle * 0.017453);
   sina = (float)sin(angle * 0.017453);

   tmp.a = 1;
   tmp.b = 0;
   tmp.c = 0;
   tmp.d = 0;
   tmp.e = cosa;
   tmp.f = -sina;
   tmp.g = 0;
   tmp.h = sina;
   tmp.i = cosa;
   tmp.j = 0;
   tmp.k = 0;
   tmp.l = 0;

   MultiplyMatrices(in, &tmp, out);
}

/*--------------------------------------------------------------*/

void TranslateMatrix(Matrix *in, Matrix *out, fpoint *pos)
{
   out->a = in->a;
   out->b = in->b;
   out->c = in->c;
   out->d = in->d;
   out->e = in->e;
   out->f = in->f;
   out->g = in->g;
   out->h = in->h;
   out->i = in->i;
   out->j = in->j + pos->x;
   out->k = in->k + pos->y;
   out->l = in->l + pos->z;
}

/*------------------------------------------------------*/
/* Point and line interaction				*/
/*------------------------------------------------------*/

/*------------------------------------------------------*/
/* line length measurement, in two dimensions		*/
/* Return the square of the length of the segment from	*/
/* userpt1 to userpt2, measuring in x and y.		*/
/*------------------------------------------------------*/

float fsqlinelen(fpoint *userpt1, fpoint *userpt2)
{
  float xdist, ydist;

  xdist = userpt2->x - userpt1->x;
  ydist = userpt2->y - userpt1->y;
  return (xdist * xdist + ydist * ydist);
}

/*------------------------------------------------------------------------*/
/* Find the closest (squared) distance from a point to a line             */
/*------------------------------------------------------------------------*/

float finddist(fpoint *linept1, fpoint *linept2, fpoint *userpt)
{
   float a, b, c, frac, protod;

   c = fsqlinelen(linept1, linept2);
   a = fsqlinelen(linept1, userpt);
   b = fsqlinelen(linept2, userpt);
   frac = a - b;
   if (frac >= c) return b;       /* "=" is important if c = 0 ! */
   else if (-frac >= c) return a;
   else {
      protod = (float)(c + a - b);
      return (a - ((protod * protod) / (c * 4)));
   }
}

#define sign(a) ((a) > 0 ? 1 : 0)

/*----------------------------------------------------------------------*/
/* Test the intersection of two line segments using winding number	*/
/* Return 1 if intersecting, 0 if not (touching returns 0)		*/
/*----------------------------------------------------------------------*/

u_char test_intersect(fpoint *seg1, fpoint *seg2)
{
   fpoint *pt1 = &seg1[0];
   fpoint *pt2 = &seg1[1];
   fpoint *pt3 = &seg2[0];
   fpoint *pt4 = &seg2[1];
   int w;
   u_char t;

   w = (pt2->x - pt1->x) * (pt3->y - pt1->y)
		- (pt2->y - pt1->y) * (pt3->x - pt1->x);
   t = sign(w);
   w = (pt3->x - pt2->x) * (pt4->y - pt2->y)
		- (pt3->y - pt2->y) * (pt4->x - pt2->x);
   t += sign(w);
   w = (pt4->x - pt3->x) * (pt1->y - pt3->y)
		- (pt4->y - pt3->y) * (pt1->x - pt3->x);
   t += sign(w);
   w = (pt1->x - pt4->x) * (pt2->y - pt4->y)
		- (pt1->y - pt4->y) * (pt2->x - pt4->x);
   t += sign(w);

   if (t != 2) return 0; 

   w = (pt2->x - pt1->x) * (pt4->y - pt1->y)
		- (pt2->y - pt1->y) * (pt4->x - pt1->x);
   t = sign(w);
   w = (pt4->x - pt2->x) * (pt3->y - pt2->y)
		- (pt4->y - pt2->y) * (pt3->x - pt2->x);
   t += sign(w);
   w = (pt3->x - pt4->x) * (pt1->y - pt4->y)
		- (pt3->y - pt4->y) * (pt1->x - pt4->x);
   t += sign(w);
   w = (pt1->x - pt3->x) * (pt2->y - pt3->y)
		- (pt1->y - pt3->y) * (pt2->x - pt3->x);
   t += sign(w);

   return (t == 2) ? 1 : 0;
}

/*------------------------------------------------------*/
