/*--------------------------------------------------------------*/
/* misc.c:							*/
/*	Miscellaneous routines---startup, GUI initialize, etc.	*/
/* Copyright (c) 2009  Tim Edwards, Open Circuit Design		*/
/*--------------------------------------------------------------*/

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>

#include "helpme.h"

Display *dpy;
Colormap cmap;
Globaldata hms;

GLXContext      grXcontext;
XVisualInfo     *grVisualInfo;

/*----------------------------------------------------------------------*/
/* High score file reading and writing					*/
/*----------------------------------------------------------------------*/

void read_high_scores()
{
   FILE *fhs;
   int i, win;

   for (i = 0; i < 10; i++) hms.high_scores[i].score = 0;

   fhs = fopen(HIGHSCORES_FILE, "r");
   if (fhs != NULL) {
      for (i = 0; i < 10; i++) {
         if (fscanf(fhs, "%d %19s %d %d %19s",
		&hms.high_scores[i].score,
		hms.high_scores[i].date,
		&hms.high_scores[i].level, &win,
		hms.high_scores[i].name) == EOF)
	    break;
	else
	   hms.high_scores[i].win = (u_char)win;
      }
      fclose(fhs);
   }
}

/*----------------------------------------------------------------------*/
/*----------------------------------------------------------------------*/

u_char check_high_score()
{
   if (hms.score == 0) return 0;
   return (hms.score >= hms.high_scores[9].score) ? 1 : 0;
}

/*----------------------------------------------------------------------*/
/*----------------------------------------------------------------------*/

void record_high_score(u_char win)
{
   int i, j;

   for (i = 0; i < 10; i++) {
      if (hms.score >= hms.high_scores[i].score) {
	 for (j = 9; j > i; j--)
	    hms.high_scores[j] = hms.high_scores[j - 1];
         hms.high_scores[i].score = hms.score;
	 hms.high_scores[i].level = hms.level;
	 hms.high_scores[i].win = win;
	 strcpy(hms.high_scores[i].name, hms.name);
	 strftime(hms.high_scores[i].date, 19, "%D", localtime(&hms.now));
	 break;
      }
   }
}

/*----------------------------------------------------------------------*/
/*----------------------------------------------------------------------*/

void save_high_scores()
{
   FILE *fhs;
   int i;

   fhs = fopen(HIGHSCORES_FILE, "w");
   if (fhs != NULL) {
      for (i = 0; i < 10; i++) {
	 if (hms.high_scores[i].score == 0) break;
         fprintf(fhs, "%d %20s %d %d %20s\n", 
			hms.high_scores[i].score,
			hms.high_scores[i].date,
			(int)hms.high_scores[i].level,
			(int)hms.high_scores[i].win,
			hms.high_scores[i].name);
      }
      fclose(fhs);
   }
}

/*----------------------------------------------------------------------*/
/* Reimplement strdup() to use Tcl_Alloc().				*/
/* Note that "strdup" is defined as "Tcl_Strdup" in helpme.h.		*/
/*----------------------------------------------------------------------*/

char *Tcl_Strdup(const char *s)
{
   char *snew;
   int slen;

   slen = 1 + strlen(s);
   snew = Tcl_Alloc(slen);
   if (snew != NULL)
      memcpy(snew, s, slen);

   return snew;
}

/*--------------------------------------------------------------*/
/* Prepare OpenGL graphics state for rendering			*/
/*--------------------------------------------------------------*/

void render_init()
{
   glXMakeCurrent(dpy, (GLXDrawable)hms.window, grXcontext);
   glDrawBuffer(GL_BACK);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   glViewport((GLsizei)0, (GLsizei)0, hms.width, hms.height);

   glEnable(GL_BLEND);
   glEnable(GL_DEPTH_TEST);
   glEnable(GL_POINT_SMOOTH);
   glEnable(GL_LINE_SMOOTH);
   glEnable(GL_POLYGON_SMOOTH);
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   glDisable(GL_COLOR_LOGIC_OP);

   /* Need to twiddle with these? */
   glScalef(1.0 / (float)(hms.width >> 1), -1.0 / (float)(hms.height >> 1), -0.001);
   glTranslated(-(hms.width >> 1), -(hms.height >> 1), 0);

   glClearColor(1.0, 1.0, 1.0, 0.0);
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   SetLineAttributes(2.0);
}

/*--------------------------------------------------------------*/
/* Copy window up from buffer					*/
/*--------------------------------------------------------------*/

void copy_window()
{
   glXSwapBuffers(dpy, (GLXDrawable)hms.window);
}

/*--------------------------------------------------------------*/

void helpme_resizearea(ClientData clientData, XEvent *eventPtr)
{
   int savewidth, saveheight;
   XEvent discard;

   /* DestroyNotify seems to come under StructureNotifyMask. . . */ 
   if (eventPtr->type == DestroyNotify) return;

   savewidth = hms.width;
   saveheight = hms.height;
   if ((dpy != NULL) && Tk_IsMapped(hms.area)) {

      hms.width = Tk_Width(hms.area);
      hms.height = Tk_Height(hms.area);
      hms.window = Tk_WindowId(hms.area);

      if (hms.width != savewidth || hms.height != saveheight) {
	 update_window();
      }

      /* Flush all resize events from the buffer */

      while (XCheckWindowEvent(dpy, hms.window, StructureNotifyMask,
			&discard) == True);
   }
}

/*--------------------------------------------------------------*/

void helpme_drawarea(ClientData clientData, XEvent *eventPtr)
{
   XEvent discard;

   Tcl_ServiceAll();

   if (!Tk_IsMapped(hms.area)) return;
   if (hms.window != (Pixmap)NULL) copy_window();
   while (XCheckWindowEvent(dpy, hms.window, ExposureMask, &discard) == True);
}


/*--------------------------------------------------------------*/
/* GUI Initialization under Tk					*/
/* First argument is the Tk path name of the drawing window.	*/
/*--------------------------------------------------------------*/

int GUI_init(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
   Tk_Window 	tkwind, tktop;
   char 	*helpmetopwin;
   int 		i, locobjc;
   XColor	scolor;

   int attributeList[] = { GLX_RGBA, GLX_DOUBLEBUFFER, None };

   tktop = Tk_MainWindow(interp);
   if (tktop == (Tk_Window)NULL) return TCL_ERROR;

   /* Check if any parameter is a Tk window name */

   locobjc = objc;
   tkwind = (Tk_Window)NULL;
   while (locobjc > 0) {
      helpmetopwin = Tcl_GetString(objv[locobjc - 1]);
      tkwind = Tk_NameToWindow(interp, helpmetopwin, tktop);
      if (tkwind != (Tk_Window)NULL)
	 break;
      locobjc--;
   }

   if (tkwind != (Tk_Window)NULL)
      hms.topwin = tkwind;
   else
      hms.topwin = tktop;

   dpy = Tk_Display(hms.topwin);
   cmap = Tk_Colormap(hms.topwin);

   /* Allocate colors */
   XAllocNamedColor(dpy, cmap, "black", &scolor, BLACK);
   XAllocNamedColor(dpy, cmap, "white", &scolor, WHITE);
   XAllocNamedColor(dpy, cmap, "red", &scolor, RED);
   XAllocNamedColor(dpy, cmap, "blue", &scolor, BLUE);
   XAllocNamedColor(dpy, cmap, "green3", &scolor, GREEN);
   XAllocNamedColor(dpy, cmap, "yellow", &scolor, YELLOW);
   XAllocNamedColor(dpy, cmap, "orange", &scolor, ORANGE);
   XAllocNamedColor(dpy, cmap, "sienna4", &scolor, BROWN);
   XAllocNamedColor(dpy, cmap, "tan", &scolor, TAN);
   XAllocNamedColor(dpy, cmap, "purple", &scolor, PURPLE);
   XAllocNamedColor(dpy, cmap, "red3", &scolor, DARK_RED);
   XAllocNamedColor(dpy, cmap, "green4", &scolor, DARK_GREEN);
   XAllocNamedColor(dpy, cmap, "blue4", &scolor, DARK_BLUE);
   XAllocNamedColor(dpy, cmap, "lightgreen", &scolor, LIGHT_GREEN);
   XAllocNamedColor(dpy, cmap, "skyblue", &scolor, LIGHT_BLUE);
   XAllocNamedColor(dpy, cmap, "gray50", &scolor, GRAY);
   XAllocNamedColor(dpy, cmap, "gray80", &scolor, LIGHT_GRAY);
   XAllocNamedColor(dpy, cmap, "gray40", &scolor, DARK_GRAY);

   /* OpenGL setup */

   grVisualInfo = glXChooseVisual(dpy, DefaultScreen(dpy), attributeList);
   grXcontext = glXCreateContext(dpy, grVisualInfo, NULL, GL_FALSE);

   return TCL_OK;
}

/*---------------------*/
/* Create a new window */
/*---------------------*/

int new_window(Tcl_Interp *interp, Tk_Window tkwind)
{
   int 		i, locobjc;
   char *helpmetopwin, *helpmedrawwin;
   char winpath[512];
   Font font;

   /* Check if this window already exists.  If so, return a pointer to it. */

   if (hms.topwin == tkwind) return -1;
   if (hms.topwin == tkwind) return -1;

   /* Before creating a new window structure, we need to make sure that	*/
   /* the target window is a "simple" window (from tkSimple.c), which 	*/
   /* we do by passing the command "<window_path> cget -commandproc"	*/
   /* and checking if it returns TCL_OK.				*/

   sprintf(winpath, "%s cget -commandproc", Tk_PathName(tkwind));
   if (Tcl_EvalEx(interp, winpath, -1, 0) != TCL_OK) {
      Tcl_SetResult(interp, "Window is the wrong type for rendering!", NULL);
      return -2;
   }

   /* Create a new window */

   hms.area = tkwind;

   /* Set initial values for the display */

   /* Make sure the window is mapped */

   Tk_MapWindow(tkwind);
   hms.window = Tk_WindowId(tkwind);
   hms.width = Tk_Width(tkwind);
   hms.height = Tk_Height(tkwind);

   /* Set up event handlers for the drawing area.			*/
   /* There are purposely no callback functions for this window---it 	*/
   /* is defined as type "simple" to keep down the cruft, as I will	*/
   /* define my own event handlers.					*/

   Tk_CreateEventHandler(tkwind, StructureNotifyMask, 
		(Tk_EventProc *)helpme_resizearea, (ClientData)NULL);
   Tk_CreateEventHandler(tkwind, ExposureMask, 
		(Tk_EventProc *)helpme_drawarea, (ClientData)NULL);

   glLineWidth(1.0);
   glShadeModel(GL_FLAT);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   glXMakeCurrent(dpy, (GLXDrawable)hms.window, grXcontext);

   /* Create the fonts */

   font = XLoadFont(dpy, "-*-times-medium-r-normal--20-*-*-*-*-*-*-*");
   LoadFont(0, font);
   hms.fonts[0] = XQueryFont(dpy, (XID)font);

   font = XLoadFont(dpy, "-*-times-bold-r-normal--34-*-*-*-*-*-*-*");
   LoadFont(1, font);
   hms.fonts[1] = XQueryFont(dpy, (XID)font);


   /* Load the texture maps */

   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

   new_texture(GRASS_TEXTURE_FILE, &hms.grass_map);
   glBindTexture(GL_TEXTURE_2D, hms.grass_map.id);
   glTexImage2D(GL_TEXTURE_2D, 0, 3, hms.grass_map.width,
	hms.grass_map.height, 0, GL_RGB, GL_UNSIGNED_BYTE,
	hms.grass_map.data);

   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

   new_texture(WOOD_TEXTURE_FILE, &hms.wood_map);
   glBindTexture(GL_TEXTURE_2D, hms.wood_map.id);
   glTexImage2D(GL_TEXTURE_2D, 0, 3, hms.wood_map.width,
	hms.wood_map.height, 0, GL_RGB, GL_UNSIGNED_BYTE,
	hms.wood_map.data);

   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

   new_texture(STONE_TEXTURE_FILE, &hms.stone_map);
   glBindTexture(GL_TEXTURE_2D, hms.stone_map.id);
   glTexImage2D(GL_TEXTURE_2D, 0, 3, hms.stone_map.width,
	hms.stone_map.height, 0, GL_RGB, GL_UNSIGNED_BYTE,
	hms.stone_map.data);

   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

   shade_texture(&hms.stone_map, &hms.stone2_map, 0.7);
   glBindTexture(GL_TEXTURE_2D, hms.stone2_map.id);
   glTexImage2D(GL_TEXTURE_2D, 0, 3, hms.stone2_map.width,
	hms.stone2_map.height, 0, GL_RGB, GL_UNSIGNED_BYTE,
	hms.stone2_map.data);

   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

   new_texture(BLUE_TEXTURE_FILE, &hms.blue_map);
   glBindTexture(GL_TEXTURE_2D, hms.blue_map.id);
   glTexImage2D(GL_TEXTURE_2D, 0, 3, hms.blue_map.width,
	hms.blue_map.height, 0, GL_RGB, GL_UNSIGNED_BYTE,
	hms.blue_map.data);

   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
   return 0;
}

/*--------------------------------------*/
/* Inline the main wrapper prodedure	*/
/*--------------------------------------*/

int helpme_start(ClientData clientData, Tcl_Interp *interp,
		int objc, Tcl_Obj *CONST objv[])
{
   char *filearg = NULL;
   int result;

   fprintf(stdout, "Starting helpme under Tcl interpreter\n");

   hms.topwin = (Tk_Window)NULL;
   hms.window = (Window)0;

   /* helpme initialization routines */

   initialize_game();
   restart_game(200, 200, 320, 10, 320);

   if (GUI_init(interp, --objc, ++objv) != TCL_OK) {
      Tcl_SetResult(interp, "Invalid or missing top-level windowname"
		" given to start command.\n", NULL);
      return TCL_ERROR;
   }

   /* All handling of files and windows is taken care of by the wrapper */

   return TCL_OK;
}

/*------------------------------------------------------*/
